#include "ActiveCalls.h"
#include "ActiveCallsInterface.h"

#define LAST_SEEN_TIMEOUT 60 /* sec */
/* how long to collect ami events */
#define AMI_SCHEDULER_TIMEOUT 100 /* ms */

#define LOG(...) ast_debug(3, __VA_ARGS__)

extern "C" {
	struct stasis_message_type *activecalls_msg_type(void);
}

class MeasureTime {
	struct timeval tv;
	const char *name;
public:

	MeasureTime(const char *name = "") : name(name)
	{
		tv = ast_tvnow();
	}

	~MeasureTime()
	{
		int64_t diff = ast_tvdiff_us(ast_tvnow(), tv);
		ast_log(AST_LOG_ERROR, "MeasureTime(%s): %"PRId64"\n", name, diff);
	}
};

/**
 * activeCallList must be locked!!!
 */
void ActiveCalls::debugPrintCalls()
{
	if (!DEBUG_ATLEAST(5)) {
		return; /* do nothing when debug log is not turned on */
	}

	BOOST_FOREACH(const ActiveCall &call, this->activeCallList)
	{
		ast_debug(5, "+++callId:%s %s(%d) => %s(%d) status:%d channels:%s dur:%d/%d %s\n",
				call.callId.c_str(),
				call.fromNumber.c_str(),
				call.fromType,
				call.toNumber.c_str(),
				call.toType,
				call.callStatus,
				call.getChannels().c_str(),
				call.computed.durationAnswered,
				call.computed.durationCreated,
				call.isConnected() ? "" : " pbxOnly"
				);
	}
}

ast_json *ActiveCalls::getCall(ActiveCalls::ActiveCallListConstIt &it)
{
	MapString mapString = boost::assign::map_list_of
			("callId", (it->callId.c_str()))
			("channels", it->getChannels().c_str())

			("fromName", it->fromName.c_str())
			("fromNumber", it->fromNumber.c_str())
			("fromNumberExternal", it->fromNumberExternal.c_str())
			("fromInterface", it->fromInterface.c_str())
			("fromCodecRead", it->computed.fromCodecRead.c_str())
			("fromCodecWrite", it->computed.fromCodecWrite.c_str())

			("toName", it->toName.c_str())
			("toNumber", it->toNumber.c_str())
			("toNumberExternal", it->toNumberExternal.c_str())
			("toInterface", it->toInterface.c_str())
			("toCodecWrite", it->computed.toCodecWrite.c_str())
			("toCodecRead", it->computed.toCodecRead.c_str())

			("transferredByNumber", it->transferredByNumber.c_str())
			("parkedByNumber", it->parkedByNumber.c_str())
			;

	MapInt mapInt = boost::assign::map_list_of
			("fromType", it->fromType)
			("toType", it->toType)
			("callStatus", (int) it->callStatus)
			("callDuration", it->computed.durationCreated)
			("callStart", it->timeCreated.tv_sec)
			("answeredDuration", it->computed.durationAnswered)
			("answeredStart", it->timeAnswered.tv_sec)
			;

	MapBool mapBool = boost::assign::map_list_of
			("isConnected", it->isConnected())
			;
	return createJson(&mapString, &mapInt, &mapBool);
}

void ActiveCalls::amiSendCalls()
{
	MutexLock lock(this->activeCallListMutex, __FILE__, __LINE__);

	LOG("Sending calls over AMI\n");
	this->compute();

	struct ast_json *msg = ast_json_object_create();
	ast_json_object_set(msg, "type", ast_json_integer_create(0 /* begin */));
	ami_send_message(activecalls_msg_type(), msg);
	ast_json_unref(msg);

	for (ActiveCallListConstIt it = this->activeCallList.begin(); it != activeCallList.end(); ++it) {
		msg = ast_json_object_create();
		ast_json_object_set(msg, "type", ast_json_integer_create(1 /* call */));
		ast_json_object_set(msg, "call", this->getCall(it));
		ami_send_message(activecalls_msg_type(), msg);
		ast_json_unref(msg);
	}

	msg = ast_json_object_create();
	ast_json_object_set(msg, "type", ast_json_integer_create(2 /* end */));
	ami_send_message(activecalls_msg_type(), msg);
	ast_json_unref(msg);
}

void ActiveCalls::deleteCall(ActiveCalls::ActiveCallListIt it)
{
	it->channels.clear();
}

/**
 * activeCallList must be locked!!!
 */
bool ActiveCalls::findChannel(ActiveCalls::ActiveCallListIt &it, const std::string &uniqueId, ActiveCalls::ChannelListIt *itChan /* = NULL */, bool *first /* = NULL */, bool *last /* = NULL */)
{
	ActiveCalls::ChannelListIt channel;

	while (it != activeCallList.end()) {
		for (channel = it->channels.begin(); channel != it->channels.end(); ++channel) {
			if (channel->uniqueId == uniqueId) {
				if (itChan) {
					*itChan = channel;
				}
				if (first) {
					*first = (channel == it->channels.begin());
				}
				if (last) {
					*last = (channel + 1 == it->channels.end());
				}
				return true;
			}
		}

		++it;
	}

	return false;
}
#define FOREACH_channel(it, uniqueId) for (ActiveCallListIt it = this->activeCallList.begin(); findChannel(it, uniqueId); ++it)

/**
 * activeCallList must be locked!!!
 */
std::string ActiveCalls::generateCallId(const std::string &callId)
{
	std::stringstream ss;
	for (int i = 1; i < 100000; i++) {
		ss.str("");
		ss << callId << "_" << i;
		if (this->activeCallList.end() == std::find_if(this->activeCallList.begin(), this->activeCallList.end(), boost::bind(&ActiveCall::callId, _1) == ss.str())) {
			break;
		}
	}
	return ss.str();
}

/**
 * activeCallList must be locked!!!
 */
void ActiveCalls::cleanup()
{
	for (ActiveCallListIt it = this->activeCallList.begin(); it != this->activeCallList.end();) {
		size_t sizeIt = it->channels.size();

		/* remove calls with empty channels */
		if (sizeIt == 0) {
			LOG("Erase call with empty channels - callId:%s\n", it->callId.c_str());
			it = this->activeCallList.erase(it);
			continue;
		}

		/* de-duplicate
		 * for example: answered ring groups */
		bool erase = false;
		for (ActiveCallListIt it2 = this->activeCallList.begin(); it2 != this->activeCallList.end(); ++it2) {
			if (it == it2) {
				continue; /* skip our self */
			}

			if (it2->channels.size() < sizeIt) {
				continue; /* nonsence to continue */
			}

			if (std::equal(it->channels.begin(), it->channels.end(), it2->channels.begin())) {
				/* it2->channels is probably longer, erase my channel */
				erase = true;
				break;
			}

			if (std::equal(it->channels.rbegin(), it->channels.rend(), it2->channels.rbegin())) {
				/* hack for feature code attended transfer */
				erase = true;
				break;
			}
		}

		if (erase) {
			LOG("Erase callId:%s to de-duplicate, channels:%s\n", it->callId.c_str(), it->getChannels().c_str());
			it = this->activeCallList.erase(it);
			continue;
		}

		++it;
	}
}

std::string ActiveCalls::getCodecDescription(const struct ast_format *format)
{
	std::string ret;

	struct ast_codec *codec = ast_format_get_codec(format);
	if (codec) {
		ret = codec->description;
		ao2_ref(codec, -1);
	}

	return ret;
}

ActiveCalls::AsteriskChannel ActiveCalls::getAsteriskChannel(struct ast_channel * c)
{
	ActiveCalls::AsteriskChannel asteriskChannel;

	ast_debug(10, "Process channel %s...\n", ast_channel_name(c));

	asteriskChannel.uniqueId = ast_channel_uniqueid(c);
	asteriskChannel.name = ast_channel_name(c);
	asteriskChannel.codecRead = getCodecDescription(ast_channel_rawreadformat(c));
	asteriskChannel.codecWrite = getCodecDescription(ast_channel_rawwriteformat(c));

	asteriskChannel.ktsMyPhoneTo = getChannelVariable("ktsMyPhoneTo", c);
	asteriskChannel.ktsCallId = getChannelVariable("ktsCallId", c);
	asteriskChannel.ktsFromNum = getChannelVariable("ktsFromNum", c);
	asteriskChannel.ktsFromName = getChannelVariable("ktsFromName", c);

	ast_debug(10, "Process channel %s... done.\n", ast_channel_name(c));
	return asteriskChannel;
}

ActiveCalls::AsteriskChannels ActiveCalls::getAsteriskChannels()
{
	AsteriskChannels asteriskChannels;
	struct ast_channel *c = NULL;
	struct ast_channel_iterator *iter = NULL;

	ast_debug(10, "Iterating all channels...\n");
	if (!(iter = ast_channel_iterator_all_new())) {
		ast_log(AST_LOG_ERROR, "Failed to get iterator for all channels.\n");
		return asteriskChannels;
	}

	ast_debug(10, "Iterating all channels... start\n");
	while ((c = ast_channel_iterator_next(iter))) {
		AsteriskChannelLock channelLock(c);
		if (channelLock.lock()) {
			AsteriskChannel asteriskChannel = this->getAsteriskChannel(c);
			channelLock.unlock();

			ast_debug(10, "Process channel %s... query t38...\n", asteriskChannel.name.c_str());
			/* do not lock any channels/bridges otherwise a deadlock with core_unreal may occur */
			if (ast_channel_get_t38_state(c) == T38_STATE_NEGOTIATED) {
				ast_debug(10, "Process channel %s... query t38... hit.\n", asteriskChannel.name.c_str());
				asteriskChannel.codecRead = "T.38";
				asteriskChannel.codecWrite = "T.38";
			}

			asteriskChannels[asteriskChannel.uniqueId] = asteriskChannel;
		}

		ast_channel_unref(c);
	}

	ast_channel_iterator_destroy(iter);
	ast_debug(10, "Iterating all channels... done\n");
	return asteriskChannels;
}

/**
 * activeCallList must be locked!!!
 */
void ActiveCalls::compute()
{
	bool runCleanup = false;
	AsteriskChannels asteriskChannels = getAsteriskChannels();

	BOOST_FOREACH(ActiveCall &activeCall, this->activeCallList)
	{
		std::string ktsCallId = activeCall.ktsCallId;

		/* compute calls */
		activeCall.compute(asteriskChannels);

		/* overwrite callid by ktsCallId
		 * used when dialing from Kerio Connect or Myphone
		 */
		if (ktsCallId != activeCall.ktsCallId) {
			std::string newCallId = generateCallId(activeCall.ktsCallId);
			LOG("Update callId:%s to %s\n", activeCall.callId.c_str(), newCallId.c_str());
			activeCall.callId = newCallId;
		}

		/* update from and to Types - this might need to go somewhere in cel() */
		if (activeCall.fromType == KTS_ExtensionTypeUnknown) {
			activeCall.fromType = this->getExtensionTypeConf(activeCall.fromNumber);
			if (activeCall.fromType == KTS_ExtensionTypeUnknown) {
				activeCall.fromType = KTS_ExtensionTypeOutboundRoute;
				LOG("Update callId:%s fromType=%d [config:unknown]\n", activeCall.callId.c_str(), activeCall.fromType);
			}
			LOG("Update callId:%s fromType=%d [config]\n", activeCall.callId.c_str(), activeCall.fromType);
		}
		if (activeCall.toType == KTS_ExtensionTypeUnknown) {
			activeCall.toType = this->getExtensionTypeConf(activeCall.toNumber);
			if (activeCall.toType == KTS_ExtensionTypeUnknown) {
				activeCall.toType = KTS_ExtensionTypeOutboundRoute;
				LOG("Update callId:%s toType=%d [config:unknown]\n", activeCall.callId.c_str(), activeCall.toType);
			}
			LOG("Update callId:%s toType=%d [config]\n", activeCall.callId.c_str(), activeCall.toType);
		}

		/* look for stalled calls */
		if (activeCall.getLastSeen() > LAST_SEEN_TIMEOUT) {
			for (std::vector<Channel>::iterator it = activeCall.channels.begin(); it != activeCall.channels.end();) {
				if (asteriskChannels.find(it->uniqueId) == asteriskChannels.end()) {
					LOG("Staled channel detected - toChannel %s was not found in asterisk, removing.\n", it->uniqueId.c_str());
					it = activeCall.channels.erase(it);
					runCleanup = true;
				}
				else {
					++it;
				}
			}
		}
	}

	/* delete stalled calls */
	if (runCleanup) {
		this->cleanup();
	}
}

ActiveCalls::ActiveCalls() : amiSchedulerId(-1), amiScheduler(NULL), activeCallListMutex("activeCallList")
{
	amiSchedulerDue = ast_tv(0, 0);

	this->reload();

	this->startAmiScheduler();
}

ActiveCalls::~ActiveCalls()
{
	if (this->amiScheduler) {
		ast_sched_context_destroy(this->amiScheduler);
		this->amiScheduler = NULL;
		ast_log(LOG_NOTICE, "Scheduler stopped\n");
	}
}

void ActiveCalls::startAmiScheduler()
{
	if (this->amiScheduler != NULL) {
		return;
	}

	if (!(this->amiScheduler = ast_sched_context_create())) {
		ast_log(LOG_ERROR, "Failed to create scheduler context\n");
		this->amiScheduler = NULL;
		return;
	}

	if (ast_sched_start_thread(this->amiScheduler)) {
		ast_sched_context_destroy(this->amiScheduler);
		this->amiScheduler = NULL;
		return;
	}
}

void ActiveCalls::reload()
{
	activeCallsExtensionList.clear();

	try {
		boost::property_tree::ptree pt;
		boost::property_tree::json_parser::read_json("/etc/asterisk/kerio_activeCalls.conf", pt);

		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_activeCalls.extensions"))
		{
			ActiveCallsExtension ext(val.second);
			activeCallsExtensionList.insert(std::pair<std::string, ActiveCallsExtension > (ext.telNum, ext));
		}
	}
	catch (std::exception & e) {
		ast_log(AST_LOG_ERROR, "Configuration error %s\n", e.what());
	}
}

void ActiveCalls::print(const int fd)
{
	MutexLock lock(this->activeCallListMutex, __FILE__, __LINE__);

	this->compute();

	ast_cli(fd, "[\n");

	for (ActiveCallListConstIt it = this->activeCallList.begin(); it != activeCallList.end(); ++it) {
		struct ast_json *jsonCall = this->getCall(it);
		char *str = ast_json_dump_string_format(jsonCall, AST_JSON_PRETTY);
		ast_json_unref(jsonCall);

		ast_cli(fd, "%s\n", str);
		ast_json_free(str);

		if (it + 1 == activeCallList.end()) {
			break; /* do not add the last comma */
		}
		ast_cli(fd, ",\n");
	}
	ast_cli(fd, "]\n");
}

void ActiveCalls::cel(struct ast_event *event)
{
	{
		MutexLock lock(this->activeCallListMutex, __FILE__, __LINE__);
		//MeasureTime measureTime("cel");

		struct ast_cel_event_record record = {
			.version = AST_CEL_EVENT_RECORD_VERSION,
		};
		MapString extraStr;
		MapInt extraInt;

		if (0 != ast_cel_fill_record(event, &record)) {
			return;
		}

		const char *event_name = record.event_name;
		if (record.event_type == AST_CEL_USER_DEFINED) {
			event_name = record.user_defined_name;
		}
		LOG("EVENT %s - unique_id:%s, linked_id:%s, extra:%s, channel:%s\n", event_name, record.unique_id, record.linked_id, record.extra, record.channel_name);

		/* update call last seen timestamp */
		FOREACH_channel(it, record.unique_id)
		{
			it->updateLastSeen();
		}

		{
			//MeasureTime measureTime(record.event_name);
			switch (record.event_type) {
				case AST_CEL_CHANNEL_START:
				{
					parseJson(record.extra, &extraStr);
					std::string requestorId = extraStr["requestor_id"];

					bool channelFirst = false;
					bool channelLast = false;
					ActiveCalls::ActiveCallListIt itCall = activeCallList.begin();
					ActiveCalls::ChannelListIt itChannel;
					findChannel(itCall, requestorId, &itChannel, &channelFirst, &channelLast);

					if (itCall == activeCallList.end()) {
						/* channel not found, create a new call */
						LOG("AST_CEL_CHANNEL_START: new call - requestor:%s unique_id:%s\n", requestorId.c_str(), record.unique_id);

						ActiveCall newCall(&record);

						activeCallList.push_back(newCall);
						itCall = activeCallList.end() - 1; /* last element */
					}
					else if (channelFirst && channelLast) {
						LOG("AST_CEL_CHANNEL_START: single, append - callId:%s\n", itCall->callId.c_str());
						itCall->appendRecord(&record);
					}
					else if (channelLast) {
						/* requestor is the last channel */
						if (itChannel->isLocal()) {
							/* requestor is the last channel and is local => append */
							LOG("AST_CEL_CHANNEL_START: last local, append - callId:%s\n", itCall->callId.c_str());

							itCall->appendRecord(&record);
						}
						else {
							/* hack for feature code attended transfer */
							/* requestor is the last channel and is not local */
							LOG("AST_CEL_CHANNEL_START: last, new call - callId:%s\n", itCall->callId.c_str());

							ActiveCall newCall(&record);
							newCall.prependLast(*itCall);
							newCall.callId = generateCallId(newCall.callId);

							activeCallList.push_back(newCall);
							itCall = activeCallList.end() - 1; /* last element */
						}
					}
					else if (itChannel->isLocal() == 1 && !boost::starts_with(record.channel_name, "Local/")) {
						/* requestor is local channel 1 and it is not local => prepend */
						LOG("AST_CEL_CHANNEL_START: prepend - callId:%s\n", itCall->callId.c_str());

						itCall->prependRecord(&record);
					}
					else {
						/* duplicate the call */
						LOG("AST_CEL_CHANNEL_START: duplicate - callId:%s\n", itCall->callId.c_str());

						ActiveCall newCall(*itCall);

						bool found = false;
						ChannelListIt itChan = newCall.channels.begin();
						while (itChan != newCall.channels.end()) {
							if (itChan->uniqueId == requestorId) {
								found = true;
								break;
							}
							++itChan;
						}

						if (!found || itChan == newCall.channels.end()) {
							return; /* should not happen */
						}

						/* erase any subsequent channels */
						newCall.channels.erase(itChan + 1, newCall.channels.end());
						newCall.appendRecord(&record);
						newCall.callId = generateCallId(newCall.callId);

						activeCallList.push_back(newCall);
						itCall = activeCallList.end() - 1; /* last element */
					}

					/* when a call is connected to another channel, mark it as not answered.
					 * solves for example dialing from myphone */
					itCall->timeAnswered.tv_sec = 0;
					itCall->timeAnswered.tv_usec = 0;
					itCall->callStatus = KTS_CallStatusConnecting;
				}
					break;
				case AST_CEL_CHANNEL_END:
				{
					/* remove the current channel */
					for (ActiveCallListIt it = this->activeCallList.begin(); it != this->activeCallList.end(); ++it) {
						it->deleteChannel(record.unique_id);
					}
				}
					break;
				case AST_CEL_LINKEDID_END:
				{
					/* Note: this event may happen before a bridge is updated,
					 * thus it is legal to have channels linked to an ended linked_id */
				}
					break;
				case AST_CEL_BLINDTRANSFER:
				{
					parseJson(record.extra, &extraStr);
					const std::string transferee = extraStr["transferee_channel_uniqueid"];
					const std::string extension = extraStr["extension"];

					/* Asterisk issue - transferee_channel_uniqueid sometimes contains invalid values
					 *
					 * 1. Call from 13 to 10, 13 did a transfer
					 *    - channels:1444818227.274(SIP/13-00000017)&1444818227.278(SIP/10-00000018)
					 *    - transferee_channel_uniqueid contains the channel that should remain in the call - 1444818227.278
					 *
					 * 2. Call from 0213 to ext. 10, findme followme to 12, picked up by 12, transfered by 12
					 *    - channels:1444819587.0(SIP/i-bender-inbound-00000000)&1444819595.14(Local/12@sip-locals-findme-00000001;1)&1444819595.15(Local/12@sip-locals-findme-00000001;2)&1444819595.23(SIP/12-00000003)
					 *    - transferee_channel_uniqueid contains the channel that did the transfer - 1444819595.15
					 *
					 * Hotfix: reverse the call, when the transferee is a local channel
					 */

					/* Update the remoteNumber - for transfers to a PBX service */
					bool channelFirst = false;
					ActiveCalls::ChannelListIt itTransferee;
					for (ActiveCalls::ActiveCallListIt itCall = activeCallList.begin(); findChannel(itCall, transferee, &itTransferee, &channelFirst); ++itCall) {
						if (itTransferee->isLocal() == 2) {
							itCall->reverse();
						}

						if (channelFirst) {
							LOG("AST_CEL_BLINDTRANSFER: update callId:%s toNumber=(%s -> %s) [record.extra.extension]\n", itCall->callId.c_str(), itCall->toNumber.c_str(), extension.c_str());
							itCall->transferredByNumber = itCall->toNumber;
							itCall->toNumber = extension;
							itCall->toName = "";
						}
						else {
							LOG("AST_CEL_BLINDTRANSFER: update callId:%s fromNumber=(%s -> %s) [record.extra.extension]\n", itCall->callId.c_str(), itCall->fromNumber.c_str(), extension.c_str());
							itCall->transferredByNumber = itCall->fromNumber;
							itCall->fromNumber = extension;
							itCall->fromName = "";

							LOG("AST_CEL_BLINDTRANSFER: reverse callId:%s\n", itCall->callId.c_str());
							itCall->reverse();
						}
					}
				}
					break;
				case AST_CEL_ATTENDEDTRANSFER:
				{
					parseJson(record.extra, &extraStr);
					const std::string transferee = extraStr["transferee_channel_uniqueid"];
					/* variable "transfer_target_channel_uniqueid" sometimes contains "N/A" thus we must use the other end of "channel2_uniqueid" const std::string target = extraStr["transfer_target_channel_uniqueid"]; */
					const std::string target = extraStr["channel2_uniqueid"];
					//bool isFeatureCodeTransfer = (extraStr["transfer_target_channel_name"] == "N/A" && extraStr["transfer_target_channel_uniqueid"] == "N/A");

					/* update source */
					ActiveCallListIt itTransferee = activeCallList.begin();
					bool transfereeFirst = false;
					if (!findChannel(itTransferee, transferee, NULL, &transfereeFirst)) {
						LOG("AST_CEL_ATTENDEDTRANSFER: transferee %s not found\n", transferee.c_str());
						return; /* not found */
					}
					if (!transfereeFirst) {
						LOG("AST_CEL_ATTENDEDTRANSFER: reverse transferee callId:%s\n", itTransferee->callId.c_str());
						itTransferee->reverse();
					}


					/* find target */
					ActiveCallListIt itTarget = activeCallList.begin();
					bool targetFirst = false;
					if (!findChannel(itTarget, target, NULL, &targetFirst)) {
						LOG("AST_CEL_ATTENDEDTRANSFER: target %s not found\n", target.c_str());
						return; /* not found */
					}

					/* feature code transfer fix */
					while (itTarget == itTransferee) {
						LOG("AST_CEL_ATTENDEDTRANSFER: target is the same as transferee callId:%s, next\n", itTarget->callId.c_str());
						itTarget++;
						if (!findChannel(itTarget, target, NULL, &targetFirst)) {
							LOG("AST_CEL_ATTENDEDTRANSFER: target %s not found\n", target.c_str());
							return; /* not found */
						}
					}

					if (!targetFirst) {
						LOG("AST_CEL_ATTENDEDTRANSFER: reverse target callId:%s\n", itTarget->callId.c_str());
						itTarget->reverse();
					}

					/* fix parked call status */
					if (itTarget->callStatus == KTS_CallStatusParked) {
						itTransferee->callStatus = itTarget->callStatus;
					}

					/* update one channel */
					LOG("AST_CEL_ATTENDEDTRANSFER: swap last channel of callId:%s by last channel of callId:%s and remove callId:%s\n", itTarget->callId.c_str(), itTransferee->callId.c_str(), itTarget->callId.c_str());

					itTransferee->transferredByNumber = itTransferee->toNumber;
					itTransferee->swapLast(*itTarget);

					/* remove the other call */
					deleteCall(itTarget);
				}
					break;
				case AST_CEL_PICKUP:
				{
					parseJson(record.extra, &extraStr);
					std::string who = extraStr["pickup_channel_uniqueid"];

					/* find who did the pickup */
					ActiveCallListIt itWho = activeCallList.begin();
					bool whoFirst = false;
					if (!findChannel(itWho, who, NULL, &whoFirst)) {
						LOG("AST_CEL_PICKUP: who %s not found\n", who.c_str());
						return; /* not found */
					}
					if (whoFirst) {
						LOG("AST_CEL_PICKUP: reverse who callId:%s\n", itWho->callId.c_str());
						itWho->reverse();
					}
					/* itWho->To is the correct leg */


					/* find target call - will be replaced */
					ActiveCallListIt itTarget = activeCallList.begin();

					if (!findChannel(itTarget, record.unique_id)) {
						LOG("AST_CEL_PICKUP: target %s not found\n", record.unique_id);
						return;
					}

					LOG("AST_CEL_PICKUP: swap channels %s and %s and delete callId:%s\n", record.unique_id, who.c_str(), itWho->callId.c_str());
					itTarget->swapLast(*itWho);
					deleteCall(itWho);

					/* in case we picked up a ring group, delete all other calls */
					FOREACH_channel(it, record.linked_id)
					{
						if (it == itTarget) {
							continue; /* skip target */
						}

						LOG("AST_CEL_PICKUP: delete callId:%s\n", it->callId.c_str());
						deleteCall(it);
					}
				}
					break;
				case AST_CEL_PARK_START:
				{
					parseJson(record.extra, &extraStr, &extraInt);
					int parkingSpace = extraInt["parking_space"];
					std::string parkingExten = extraStr["parking_exten"];

					ActiveCallListIt it = activeCallList.begin();
					bool first = false, last = false;
					if (!findChannel(it, record.unique_id, NULL, &first, &last)) {
						LOG("AST_CEL_PARK_START: channel %s not found\n", record.unique_id);
						return;
					}
					it->callStatus = KTS_CallStatusParked;

					/* reconstruct the full parking service number,
					 * useful when parking using the auto position e.g. *5 => *500 */
					std::string parkingNumber;
					char parkingSpace2[64];
					snprintf(parkingSpace2, sizeof (parkingSpace2), "%d", parkingSpace);
					if (strlen(parkingSpace2) <= 1) {
						LOG("AST_CEL_PARK_START: Invalid parking space \"%d\"\n", parkingSpace);
						return;
					}

					const char *parkingSpaceStripped = parkingSpace2 + 1;
					parkingNumber = parkingExten + parkingSpaceStripped;
					LOG("AST_CEL_PARK_START: first:%d, last:%d, parking number %s\n", first, last, parkingNumber.c_str());

					if (!parkingNumber.empty()) {
						it->parkedByNumber = it->toNumber;
						it->toNumber = parkingNumber;
					}

					it->parkedByType = it->toType;
					it->toType = KTS_ExtensionTypeCallParkingService;
				}
					break;
				case AST_CEL_PARK_END:
				{
					parseJson(record.extra, &extraStr);
					const std::string retriever = extraStr["retriever"];
					const std::string reason = extraStr["reason"];

					/* find call */
					ActiveCallListIt itTarget = activeCallList.begin();
					if (!findChannel(itTarget, record.unique_id)) {
						LOG("AST_CEL_PARK_END: target %s not found\n", record.unique_id);
						return;
					}

					/* handle timeout */
					if (reason == "ParkedCallTimeOut") {
						itTarget->toType = itTarget->parkedByType;
						itTarget->toNumber = itTarget->parkedByNumber;
					}
					if (reason == "ParkedCallSwap") {
						/* occurs when parking using ##*5# */
						/* fix parkedByNumber */
						itTarget->parkedByNumber = itTarget->transferredByNumber;
						break;
					}

					if (retriever.size()) {
						ActiveCallListIt it = this->activeCallList.begin();
						ChannelListIt channelIt;
						for (; it != this->activeCallList.end(); ++it) {
							for (channelIt = it->channels.begin(); channelIt != it->channels.end(); ++channelIt) {
								if (channelIt->channelName == retriever) {
									break;
								}
							}

							if (channelIt != it->channels.end()) {
								break;
							}
						}

						if (it == this->activeCallList.end()) {
							LOG("AST_CEL_PARK_END: retriever %s not found\n", record.unique_id);
							return;
						}

						it->reverse();
						itTarget->appendCall(*it);
						deleteCall(it);
					}

					itTarget->callStatus = KTS_CallStatusConnected;
					itTarget->parkedByNumber = "";
					itTarget->parkedByType = 0;
				}
					break;
				case AST_CEL_ANSWER:
				{

					FOREACH_channel(it, record.unique_id)
					{
						it->timeAnswered = ast_tvmin(it->timeAnswered, record.event_time);

						switch (it->callStatus) {
							case KTS_CallStatusConnected:
							case KTS_CallStatusConnecting:
							case KTS_CallStatusPickUp:
								it->callStatus = KTS_CallStatusConnected;
								break;
							default:
								break;
						}
					}
				}
					break;
				case AST_CEL_USER_DEFINED:
				{
					parseJson(record.extra, &extraStr);

					FOREACH_channel(it, record.unique_id)
					{
						/* DISABLED, breaks transfers to outbound calls
						 * if (!extraStr["ktsFrom"].empty()) {
						 *	it->fromNumber = extraStr["ktsFrom"];
						 * }
						 */

						/* Variable ktsFrom must be used for analog calls. These calls learn callerid between the first&second ring, not during CHAN_START
						 * Fixes: OPERATOR-4993 - Active calls: Invalid numbers with analogue interfaces
						 * Use ktsFrom only when ktsForce = 1 */
						if (extraStr["ktsForce"] == "1" && !extraStr["ktsFrom"].empty()) {
							it->fromNumber = extraStr["ktsFrom"];
						}
						if (!extraStr["ktsTo"].empty()) {
							it->toNumber = extraStr["ktsTo"];
						}
						if (!extraStr["ktsFromExternal"].empty()) {
							it->fromNumberExternal = extraStr["ktsFromExternal"];
						}
						if (!extraStr["ktsToExternal"].empty()) {
							it->toNumberExternal = extraStr["ktsToExternal"];
						}
						if (!extraStr["ktsFromInterface"].empty()) {
							it->fromInterface = extraStr["ktsFromInterface"];
						}
						if (!extraStr["ktsToInterface"].empty()) {
							it->toInterface = extraStr["ktsToInterface"];
						}
						if (!extraStr["ktsFromType"].empty()) {
							it->fromType = getExtensionType(extraStr["ktsFromType"]);
						}
						if (!extraStr["ktsToType"].empty()) {
							it->toType = getExtensionType(extraStr["ktsToType"]);
						}

						if (!extraStr["ktsStatus"].empty()) {
							if (extraStr["ktsStatus"] == "Voicemail") {
								it->callStatus = KTS_CallStatusVoicemail;
								it->timeAnswered = ast_tvmin(it->timeAnswered, record.event_time); /* make calls answered by voicemail answered */
							}
							if (extraStr["ktsStatus"] == "Followme") {
								it->callStatus = KTS_CallStatusFollowme;
							}
							if (extraStr["ktsStatus"] == "ReceiveFax") {
								it->callStatus = KTS_CallStatusReceiveFax;
							}
							if (extraStr["ktsStatus"] == "SendFax") {
								it->callStatus = KTS_CallStatusSendFax;
							}
							if (extraStr["ktsStatus"] == "Busy") {
								it->callStatus = KTS_CallStatusBusy;
							}
							if (extraStr["ktsStatus"] == "Forbidden") {
								it->callStatus = KTS_CallStatusForbidden;
							}
							if (extraStr["ktsStatus"] == "Blacklisted") {
								it->callStatus = KTS_CallStatusBlacklisted;
							}
						}
					}
				}
					break;
				default:
					break;
			}
		}

		{
			//MeasureTime measureTime("cleanup");
			switch (record.event_type) {
				case AST_CEL_CHANNEL_START:
				case AST_CEL_CHANNEL_END:
				case AST_CEL_BLINDTRANSFER:
				case AST_CEL_ATTENDEDTRANSFER:
				case AST_CEL_PICKUP:
				case AST_CEL_PARK_START:
				case AST_CEL_PARK_END:
					this->cleanup();
					break;
				default:
					break;
			}

			/* after each change print channels and calls to the log */
			this->debugPrintCalls();
		}
	} /* end of mutex, amiScheduler may deadlock otherwise */

	{
		//MeasureTime measureTime("amiScheduler");

		/* Send an ami message - asynchronous using the asterisk's scheduler */
		struct timeval now = ast_tvnow();
		int64_t timeLeft = ast_tvdiff_ms(this->amiSchedulerDue, now);
		if (timeLeft > 0 && timeLeft <= AMI_SCHEDULER_TIMEOUT) {
			/* we're still within amiSchedulerDue */
			LOG("AmiScheduler - Try to reschedule.\n");

			/* re-schedule, delete */
			if (this->amiSchedulerId > -1) {
				int _sched_res = ast_sched_del(this->amiScheduler, this->amiSchedulerId);
				if (_sched_res < 0) {
					LOG("AmiScheduler - Use a new scheduler (deleted).\n");
					this->amiSchedulerId = -1;
				}
			}
		}
		else {
			/* re-schedule */
			LOG("AmiScheduler - Use a new scheduler (out of timeout).\n");
			this->amiSchedulerId = -1;
		}

		if (this->amiSchedulerId == -1) {
			this->amiSchedulerDue = ast_tvadd(now, ast_tv(0, AMI_SCHEDULER_TIMEOUT * 1000));
			timeLeft = AMI_SCHEDULER_TIMEOUT; /* ms */
		}

		this->amiSchedulerId = ast_sched_add(this->amiScheduler, (int) timeLeft, activecalls_amiSendCallsCb, 0);
		LOG("AmiScheduler - New scheduler %d created with timeout %d ms.\n", this->amiSchedulerId, (int) timeLeft);
	}
}

KTS_ExtensionType ActiveCalls::getExtensionTypeConf(const std::string &telNum)
{
	ActiveCallsExtensionList::iterator it = activeCallsExtensionList.find(telNum);
	if (it != activeCallsExtensionList.end()) {
		return (KTS_ExtensionType) it->second.extensionTypeId;
	}
	return KTS_ExtensionTypeUnknown;
}

KTS_ExtensionType ActiveCalls::getExtensionType(const std::string &value)
{
	if ("Line" == value) {
		return KTS_ExtensionTypeLine;
	}
	else if ("Interface" == value) {
		return KTS_ExtensionTypeInterface;
	}
	else if ("OutboundRoute" == value) {
		return KTS_ExtensionTypeOutboundRoute;
	}
	else if ("Conference" == value) {
		return KTS_ExtensionTypeConference;
	}
	else if ("CallQueue" == value) {
		return KTS_ExtensionTypeCallQueue;
	}
	else if ("RingGroup" == value) {
		/* present ring group calls as extension calls: return KTS_ExtensionTypeRingGroup; */
		return KTS_ExtensionTypeLine;
	}
	else if ("PagingGroup" == value) {
		return KTS_ExtensionTypePagingGroup;
	}
	else if ("PagingService" == value) {
		return KTS_ExtensionTypePagingService;
	}
	else if ("Ivr" == value) {
		return KTS_ExtensionTypeIvr;
	}
	else if ("Voicemail" == value) {
		return KTS_ExtensionTypeVoicemail;
	}
	else if ("EmergencyNumber" == value) {
		return KTS_ExtensionTypeEmergencyNumber;
	}
	else if ("EchoService" == value) {
		return KTS_ExtensionTypeEchoService;
	}
	else if ("MusicService" == value) {
		return KTS_ExtensionTypeMusicService;
	}
	else if ("DialByExtService" == value) {
		return KTS_ExtensionTypeDialByExtService;
	}
	else if ("DialByNameService" == value) {
		return KTS_ExtensionTypeDialByNameService;
	}
	else if ("CurrentTimeService" == value) {
		return KTS_ExtensionTypeCurrentTimeService;
	}
	else if ("RecordAudioService" == value) {
		return KTS_ExtensionTypeRecordAudioService;
	}
	else if ("CallPickupService" == value) {
		return KTS_ExtensionTypeCallPickupService;
	}
	else if ("DirectedCallPickupService" == value) {
		return KTS_ExtensionTypeDirectedCallPickupService;
	}
	else if ("SpeedDial" == value) {
		return KTS_ExtensionTypeSpeedDial;
	}
	else if ("CallParkingService" == value) {
		return KTS_ExtensionTypeCallParkingService;
	}
	else if ("CallParkingParkService" == value) {
		return KTS_ExtensionTypeCallParkingParkService;
	}
	else if ("CallParkingUnparkService" == value) {
		return KTS_ExtensionTypeCallParkingUnparkService;
	}
	else if ("FaxReceiveService" == value) {
		return KTS_ExtensionTypeFaxReceiveService;
	}
	else if ("SpyService" == value) {
		return KTS_ExtensionTypeSpyService;
	}
	else if ("DayNightModeService" == value) {
		return KTS_ExtensionTypeDayNightModeService;
	}
	else if ("BlacklistAddService" == value) {
		return KTS_ExtensionTypeBlacklistAddService;
	}
	else if ("BlacklistAddLastService" == value) {
		return KTS_ExtensionTypeBlacklistAddLastService;
	}
	else if ("BlacklistRemoveService" == value) {
		return KTS_ExtensionTypeBlacklistRemoveService;
	}
	else {
		ast_log(AST_LOG_WARNING, "Unknown KTS_ExtensionType \"%s\"\n", value.c_str());
		return KTS_ExtensionTypeUnknown;
	}
}

std::string ActiveCalls::getChannelVariable(const std::string paramName, struct ast_channel *chan, const std::string defaultValue /* = "" */)
{
	const char *var;
	std::string paramValue;

	ast_channel_lock(chan);
	if ((var = pbx_builtin_getvar_helper(chan, paramName.c_str()))) {
		paramValue = var;
	}
	ast_channel_unlock(chan);

	return paramValue.empty() ? defaultValue : paramValue;
}

void ActiveCalls::setChannelVariable(struct ast_channel *chan, const std::string paramName, const std::string paramValue)
{
	pbx_builtin_setvar_helper(chan, paramName.c_str(), paramValue.c_str());
}

std::string ActiveCalls::getCdrVariable(const std::string &name, const std::string &channelName, const std::string defaultValue /* = "" */)
{
	char buf[1024];
	/* ast_cdr_getvar is case insensitive */
	if (0 == ast_cdr_getvar(channelName.c_str(), name.c_str(), buf, sizeof (buf))) {
		return std::string(buf);
	}

	return defaultValue;
}

/**
 * Hangup outgoing calls
 *
 * @return [int] - number of hang up calls
 */
int ActiveCalls::hangupOutgoingCalls(std::vector<std::string> emergencyList, std::vector<std::string> routePrefixes)
{
	MutexLock lock(this->activeCallListMutex, __FILE__, __LINE__);

	int count = 0;

	/* compute durations and toType */
	this->compute();

	BOOST_FOREACH(const ActiveCall &call, this->activeCallList)
	{
		/* outbound only */
		if (KTS_ExtensionTypeOutboundRoute != call.toType) {
			continue;
		}

		/* skip emergency numbers */
		bool isEmergency = false;
		std::string calledNumber = call.toNumber;

		BOOST_FOREACH(std::string routePrefix, routePrefixes)
		{
			if (isEmergency) {
				break;
			}

			BOOST_FOREACH(std::string emergencyNumber, emergencyList)
			{
				LOG("Search %s%s == %s\n", routePrefix.c_str(), emergencyNumber.c_str(), calledNumber.c_str());
				if (routePrefix + emergencyNumber == calledNumber) {
					isEmergency = true;
					break;
				}
			}
		}

		if (isEmergency) {
			ast_log(AST_LOG_NOTICE, "Not hanging up emergency call to %s from callid:%s.\n", calledNumber.c_str(), call.callId.c_str());
			continue;
		}

		/* channel can't be empty */
		if (call.channels.size() == 0) {
			continue;
		}

		/* HANG UP */
		struct ast_channel *c = NULL;
		struct ast_channel_iterator *iter;
		if (!(iter = ast_channel_iterator_by_name_new(call.channels[0].channelName.c_str(), call.channels[0].channelName.size()))) {
			continue;
		}

		while ((c = ast_channel_iterator_next(iter))) {
			ast_channel_lock(c);
			ast_log(AST_LOG_WARNING, "Soft hanging %s up.\n", ast_channel_name(c));
			ast_softhangup(c, AST_SOFTHANGUP_EXPLICIT);
			ast_channel_unlock(c);
			c = ast_channel_unref(c);
			count++;
		}

		ast_channel_iterator_destroy(iter);
	}

	return count;
}

/**
 * Get calls duration
 *
 * @param scope
 * @param prefixes
 * @return [int] - duration in seconds
 */
int ActiveCalls::getCallsDurationPrefix(int scope, std::vector<std::string> prefixes)
{
	MutexLock lock(this->activeCallListMutex, __FILE__, __LINE__);

	int durationSeconds = 0;

	/* compute durations and toType */
	this->compute();

	BOOST_FOREACH(const ActiveCall &call, this->activeCallList)
	{
		/* outbound only */
		if (KTS_ExtensionTypeOutboundRoute != call.toType) {
			continue;
		}
		/* answered only */
		if (!call.isConnected()) {
			continue;
		}

		switch (scope) {
			case 0: /* all */
				durationSeconds += call.computed.durationAnswered / 1000;
				break;
			case 1: /* only */
			{

				BOOST_FOREACH(const std::string &prefix, prefixes)
				{
					if (0 == call.toNumber.find(prefix)) {
						durationSeconds += call.computed.durationAnswered / 1000;
						break;
					}
				}
				break;
			}
			case 2: /* except */
			{
				bool found = false;

				BOOST_FOREACH(const std::string &prefix, prefixes)
				{
					if (0 == call.toNumber.find(prefix)) {
						found = true;
						break;
					}
				}
				if (!found) {
					durationSeconds += call.computed.durationAnswered / 1000;
				}
				break;
			}
		}
	}

	return durationSeconds;
}

/**
 * Get calls count
 *
 * @param scope
 * @param prefixes
 * @return [int] - duration in seconds
 */
int ActiveCalls::getCallsCountPrefix(int scope, std::vector<std::string> prefixes)
{
	MutexLock lock(this->activeCallListMutex, __FILE__, __LINE__);

	int count = 0;

	/* compute durations and toType */
	this->compute();

	BOOST_FOREACH(const ActiveCall &call, this->activeCallList)
	{
		/* outbound only */
		if (KTS_ExtensionTypeOutboundRoute != call.toType) {
			continue;
		}
		/* answered only */
		if (!call.isConnected()) {
			continue;
		}

		switch (scope) {
			case 0: /* all */
				count++;
				break;
			case 1: /* only */
			{

				BOOST_FOREACH(const std::string &prefix, prefixes)
				{
					if (0 == call.toNumber.find(prefix)) {
						count++;
						break;
					}
				}
				break;
			}
			case 2: /* except */
			{
				bool found = false;

				BOOST_FOREACH(const std::string &prefix, prefixes)
				{
					if (0 == call.toNumber.find(prefix)) {
						found = true;
						break;
					}
				}
				if (!found) {
					count++;
				}

				break;
			}
		}
	}

	return count;
}
