#include "CallConstraints.h"
#include "ActiveCallsInterface.h"
#include "CallHistoryInterface.h"
#include "ActiveCallsInterface.h"

#define fileBlockAllCalls "/etc/asterisk/kerio_callConstraintsBlockAllCalls.conf"
#define callConstraintsHelper "/opt/kerio/operator/bin/callConstraintsHelper"

CallConstraints::CallConstraints()
{
	loadConfig();
}

CallConstraints::~CallConstraints() {
 }

/**
 * Loads settings from database
 */
void CallConstraints::loadConfig(void)
{
	constraintList.clear();

	try {
		boost::property_tree::ptree pt;
		boost::property_tree::json_parser::read_json("/etc/asterisk/kerio_callConstraints.conf", pt);

		/* get constraints */
		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_callConstraints.callConstraints"))
		{
			CallConstraintsRule constr(val.second);
			constraintList.push_back(constr);
		}

		/* get lastUnblockTime */
		lastUnblockTime = pt.get<int >("kerio_callConstraints.lastUnblockTime");

		/* get prefixes */
		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_callConstraints.prefixes"))
		{
			routePrefixList.push_back(val.second.data());
		}

		/* get emergency numbers */
		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_callConstraints.emergencyNumbers"))
		{
			emergencyNumberList.push_back(val.second.data());
		}
	}
	catch (std::exception & e) {
		ast_log(AST_LOG_ERROR, "Configuration error %s\n", e.what());
	}
}

/**
 * Print constraints
 *
 * @oaram fd [int] - file descriptor for output
 */
void CallConstraints::print(int fd)
{
	ast_cli(fd, "Constrains\n");
	ast_cli(fd, "----------\n");
	ast_cli(fd, "Outgoing calls are: %s\n", getBlockAllCalls() ? "BLOCKED" : "ALLOWED");
	ast_cli(fd, "Route prefixes are:");

	BOOST_FOREACH(std::string prefix, routePrefixList)
	{
		ast_cli(fd, " '%s'", prefix.c_str());
	}
	ast_cli(fd, "\n");

	ast_cli(fd, "Emergency numbers are:");

	BOOST_FOREACH(std::string emergencyNumber, emergencyNumberList)
	{
		ast_cli(fd, " %s", emergencyNumber.c_str());
	}
	ast_cli(fd, "\n");

	ast_cli(fd, "\n");
	ast_cli(fd, "Constraint list\n");
	ast_cli(fd, "---------------\n");

	BOOST_FOREACH(CallConstraintsRule c, constraintList)
	{
		ast_cli(fd, " %15s:\n", c.name.c_str());
		if (c.callCountEnabled) {
			ast_cli(fd, "\t\t  count limit: enabled (%d/%d) over the last %d seconds %s\n", getCallsCount(c), c.callCount, c.callCountPeriod, c.callCountActive ? "- TRIGGERED" : "");
		}
		else {
			ast_cli(fd, "\t\t  count limit: disabled\n");
		}

		if (c.callDurationEnabled) {
			ast_cli(fd, "\t\t  duration limit: enabled (%d/%d seconds) over the last %d seconds %s\n", getCallsDuration(c), c.callDuration, c.callDurationPeriod, c.callDurationActive ? "- TRIGGERED" : "");
		}
		else {
			ast_cli(fd, "\t\t  duration limit: disabled\n");
		}


		if (c.actionEmail) {
			ast_cli(fd, "\t\t  email notification: enabled\n");
		}
		else {
			ast_cli(fd, "\t\t  email notification: disabled\n");
		}

		if (c.actionBlock) {
			ast_cli(fd, "\t\t  call blocking: enabled\n");
		}
		else {
			ast_cli(fd, "\t\t  call blocking: disabled\n");
		}

		std::string prefixes;

		BOOST_FOREACH(std::string prefix, c.prefixes)
		{
			prefixes += prefix + " ";
		}

		switch (c.applyToType) {
			case 0:
				ast_cli(fd, "\t\t  scope: all calls\n");
				break;
			case 1:
				ast_cli(fd, "\t\t  scope: only %s\n", prefixes.c_str());
				break;
			case 2:
				ast_cli(fd, "\t\t  scope: except %s\n", prefixes.c_str());
				break;
		}
	}
}

/**
 * Print constraints
 *
 * @oaram fd [int] - file descriptor for output
 */
void CallConstraints::printConcise(int fd)
{
	ast_cli(fd, "Constraint list\n");

	BOOST_FOREACH(CallConstraintsRule c, constraintList)
	{
		ast_cli(fd, "%d!%d!%d\n",
				c.id,
				getCallsCount(c),
				getCallsDuration(c)
				);
	}
}

/**
 * Test if call constraint applies
 *
 * @param c [CallConstraint] - call constraint
 * @param dialedNumber [std::string] - dialed number
 * @return [bool]
 */
bool CallConstraints::ruleApplies(CallConstraintsRule c, std::string dialedNumber)
{
	bool result = false;

	switch (c.applyToType) {
		case 0: /* all */
			result = true;
			break;
		case 1: /* only */
			result = false;

			BOOST_FOREACH(std::string prefix, c.prefixes)
			{
				if (0 == dialedNumber.find(prefix)) {
					result = true;
					break;
				}
			}
			break;
		case 2: /* except */
			result = true;

			BOOST_FOREACH(std::string prefix, c.prefixes)
			{
				if (0 == dialedNumber.find(prefix)) {
					result = false;
					break;
				}
			}
			break;
	}

	return result;
}

/**
 * Get count of calls for a constraint from call history (cdr) + active calls (asterisk)
 *
 * @param c [CallConstraint]
 * @return [int] - number of calls
 */
int CallConstraints::getCallsCount(CallConstraintsRule c)
{
	int count = 0;
	int timeframe = (time(NULL) - c.callCountPeriod);
	if (lastUnblockTime > timeframe) {
		timeframe = lastUnblockTime;
	}

	count = history_getCountPrefix(c.applyToType, c.prefixes, timeframe);
	count += activecalls_getCountPrefix(c.applyToType, c.prefixes);

	return count;
}

/**
 * Get duration of calls from history (cdr) + active calls (asterisk)
 * @param c [CallConstraint]
 * @return [int] - time in seconds
 */
int CallConstraints::getCallsDuration(CallConstraintsRule c)
{
	int duration = 0;
	int timeframe = (time(NULL) - c.callDurationPeriod);
	if (lastUnblockTime > timeframe) {
		timeframe = lastUnblockTime;
	}
	//ast_verb(3, "time=%ld time-callDurationPeriod=%ld lastUnblock=%d timeframe=%d\n", time(NULL), (time(NULL) - c.callDurationPeriod), lastUnblockTime, timeframe);

	duration = history_getDurationPrefix(c.applyToType, c.prefixes, timeframe);
	duration += activecalls_getDurationPrefix(c.applyToType, c.prefixes);

	return duration;
}

/**
 * Read blockAllCalls from database
 * @return true when calls are blocked
 */
bool CallConstraints::getBlockAllCalls()
{
	struct stat sb;
	if (stat(fileBlockAllCalls, &sb) == -1) {
		/* file doesn't exist */
		return false; /* ALLOWED */
	}

	return true; /* BLOCKED */
}

/**
 * Save blockAllCalls to database
 * @param value [bool] - true when calls should be blocked
 */
void CallConstraints::setBlockAllOutgoingCalls(bool value)
{
	if (value) {
		/* block */
		FILE *fp = fopen(fileBlockAllCalls, "w");
		if (!fp) {
			ast_log(AST_LOG_ERROR, "Cannot create file %s\n", fileBlockAllCalls);
			return;
		}

		fprintf(fp, "Outgoing calls are blocked. To unblock, delete this file.\n");
		fclose(fp);
	}
	else {
		/* unblock */
		unlink(fileBlockAllCalls);
	}
}

/**
 * Hangup all outgoing calls
 */
void CallConstraints::hangupOutgoingCalls(void)
{
	ast_log(AST_LOG_WARNING, "[security] Hanging up all outgoing calls except emergency...\n");
	int callCount = 0;
	callCount = activecalls_hangupOutgoingCalls(emergencyNumberList, routePrefixList);
	ast_log(AST_LOG_WARNING, "[security] Hanging up all outgoing calls except emergency... %d calls hung up\n", callCount);
}

/**
 * Send email notification
 *
 * @param c [CallConstraint]
 */
void CallConstraints::sendEmailNotification(CallConstraintsRule c)
{
	ast_log(AST_LOG_NOTICE, "Sending email notification (call constranstraints)...\n");

	char tmp[1024];
	snprintf(tmp, sizeof (tmp), "%s callConstraintTriggered "\
			"--CONSTRAINT_NAME \"%s\" "\
			"--ACTION_BLOCK %d "\
			"--CALL_COUNT_ACTIVE %d "\
			"--CALL_COUNT %d "\
			"--CALL_DURATION_ACTIVE %d "\
			"--CALL_DURATION %d &",
			mailCommand,
			c.name.c_str(),
			c.actionBlock ? 1 : 0,
			c.callCountActive ? 1 : 0,
			c.callCount,
			c.callDurationActive ? 1 : 0,
			c.callDuration
			);

	ast_safe_system(tmp);

	ast_log(AST_LOG_NOTICE, "Sending email notification (call constranstraints)... done.\n");
}

/**
 * Check if call is emergency
 *
 * @param dialedNumber
 * @return [bool] - true on emergency
 */
bool CallConstraints::callIsEmergency(std::string dialedNumber)
{

	BOOST_FOREACH(std::string routePrefix, routePrefixList)
	{

		BOOST_FOREACH(std::string emergencyNumber, emergencyNumberList)
		{
			if (routePrefix + emergencyNumber == dialedNumber) {
				return true;
			}
		}
	}
	return false;
}

/**
 * Test outgoing call constraints
 *
 * @param dialedNumber [std::string] - dialed number
 * @return [int] - 0 on deny, 1 on permit
 */
int CallConstraints::test(std::string dialedNumber)
{
	bool blockCalls = false;

	/* test if outgoing calls are blocked */
	if (getBlockAllCalls()) {
		/* emergency calls are permitted */
		if (callIsEmergency(dialedNumber)) {
			ast_log(AST_LOG_NOTICE, "Outgoing calls are blocked, but emergency call to %s is permitted...\n", dialedNumber.c_str());
			return 1;
		}

		ast_log(AST_LOG_NOTICE, "Outgoing calls are blocked, call to %s denied...\n", dialedNumber.c_str());
		return 0; /* deny call */
	}

	/* test constraints */
	BOOST_FOREACH(CallConstraintsRule &c, constraintList)
	{
		/* skip disabled conditions */
		if (!c.callCountEnabled) {
			continue;
		}

		/* get call count for the constraint */
		int callCount = getCallsCount(c);
		bool callCountActive = callCount >= c.callCount;

		/* on change */
		if (callCountActive != c.callCountActive) {
			c.callCountActive = callCountActive;
			saveConstraint(c);

			/* trigger some action on activation */
			if (callCountActive) {
				/* email */
				if (c.actionEmail) {
					if (!c.actionBlock) {
						/* remove duplicit log messages */
						ast_log(AST_LOG_WARNING, "[security] Security constraint %s has been triggered. Call limit of %d calls reached.\n", c.name.c_str(), c.callCount);
					}
					sendEmailNotification(c);
				}

				/* call blocking */
				if (c.actionBlock) {
					ast_log(AST_LOG_WARNING, "[security] Security constraint %s has been triggered. Call limit of %d calls reached. Blocking all outgoing calls.\n", c.name.c_str(), c.callCount);
					setBlockAllOutgoingCalls(true);
					blockCalls = true;
				}
			}
			else {
				ast_log(AST_LOG_NOTICE, "Security constraint %s is not active anymore. Number of calls dropped to %d/%d.\n", c.name.c_str(), callCount, c.callCount);
			}
		}

		ast_verb(3, "Current calls according to constraint %s: %d/%d.\n", c.name.c_str(), callCount, c.callCount);
	}
	return blockCalls ? 0 : 1;
}

/**
 * Checking call duration
 */
void CallConstraints::schedCheck()
{
	ast_verb(3, "Action schedCheck...\n");

	if (this->getBlockAllCalls()) {
		return; /* when outgoing calls are blocked do nothing */
	}

	/* test constraints */
	BOOST_FOREACH(CallConstraintsRule &c, this->constraintList)
	{
		/* skip disabled conditions */
		if (!c.callDurationEnabled) {
			continue;
		}

		/* get call duration for the constraint*/
		int currentDuration = this->getCallsDuration(c);
		bool callDurationActive = currentDuration > c.callDuration;

		/* on change */
		if (callDurationActive != c.callDurationActive) {
			c.callDurationActive = callDurationActive;
			this->saveConstraint(c);

			/* trigger some action on activation */
			if (callDurationActive) {
				/* email */
				if (c.actionEmail) {
					if (!c.actionBlock) {
						/* remove duplicit log messages*/
						ast_log(AST_LOG_WARNING, "[security] Security constraint %s has been triggered. Maximum call duration (%d seconds) reached.\n", c.name.c_str(), c.callDuration);
					}
					this->sendEmailNotification(c);
				}

				/* call blocking */
				if (c.actionBlock) {
					ast_log(AST_LOG_WARNING, "[security] Security constraint %s has been triggered. Maximum call duration (%d seconds) reached. Blocking all outgoing calls.\n", c.name.c_str(), c.callDuration);
					this->setBlockAllOutgoingCalls(true);
					this->hangupOutgoingCalls();
				}
			}
			else {
				ast_log(AST_LOG_NOTICE, "Security constraint %s is not active anymore. Total call duration dropped to %d/%d seconds.\n", c.name.c_str(), currentDuration, c.callDuration);
			}
		}
		ast_debug(2, "Current call duration according to constraint %s: %d/%d.\n", c.name.c_str(), currentDuration, c.callDuration);
	}

	ast_verb(3, "Action schedCheck... done.\n");
}

/**
 * Save MAX_COUNT_ACTIVE and MAX_TOTAL_DURATION_ACTIVE to the database
 *
 * @param c [CallConstraint]
 */
void CallConstraints::saveConstraint(CallConstraintsRule c)
{
	char cmd[1024];
	snprintf(cmd, sizeof (cmd), "%s --saveConstraint %d %d %d",
			callConstraintsHelper,
			c.callCountActive ? 1 : 0,
			c.callDurationActive ? 1 : 0,
			c.id
			);

	std::string output;
	shellExec(cmd, output);
}
