#include <boostWrapper.h>
#include <boost/algorithm/string.hpp>
#include <boost/regex.hpp>

#include "ActiveCallsInterface.h"
#include "ActiveCalls.h"
#include "CallHistory.h"

#define SQL_CDR "CREATE TABLE cdr ("\
"Guid				INTEGER PRIMARY KEY,"\
"FromNum			VARCHAR(80)," /* phone number */\
"FromName			VARCHAR(80)," /* caller id - name */\
"FromChannel		VARCHAR(80)," /* channel */\
"FromType			INTEGER," /* ExtensionTypeEnum - from db */\
"ToNum				VARCHAR(80)," /* phone number */\
"ToChannel			VARCHAR(80)," /* channel */\
"ToType				INTEGER," /* ExtensionTypeEnum - from db */\
"Status				INTEGER," /* CallStatusEnum */\
"External			VARCHAR(16)," /* External call - <in|out|> */\
"AccountCode		VARCHAR(20)," /* Account code from cdr */\
"CallInitialized	INTEGER," /* Timestamp */\
"CallDuration		INTEGER," /* seconds */\
"AnsweredDuration	INTEGER, " /* seconds */\
"UniqueId			VARCHAR(80)," /* Asterisk's uniqueId */\
"FromInterface		VARCHAR(80)," /* External interface */\
"ToInterface		VARCHAR(80)," /* External interface */\
"FromExternal		VARCHAR(80)," /* From number before it was rewritten by inbound rules */\
"ToExternal			VARCHAR(80)," /* To number before it was rewritten by inbound rules */\
"FromPeerIp			VARCHAR(16)," /* IP address (behind NAT) */\
"ToPeerIp			VARCHAR(16)," /* IP address (behind NAT) */\
"FromRecvIp			VARCHAR(16)," /* public IP address */\
"ToRecvIp			VARCHAR(16)," /* public IP address */\
"FromUserAgent		VARCHAR(80)," /* user agent */\
"ToUserAgent		VARCHAR(80)," /* user agent */\
"FromCodec			VARCHAR(16)," /* codec */\
"ToCodec			VARCHAR(16)," /* codec */\
"FromQos			VARCHAR(80)," /* qos */\
"ToQos				VARCHAR(80)," /* qos */\
"Details			VARCHAR(100)" /* more details, e.g. call queue agent, ivr dial number, ring group member */\
");"

extern "C" {
	struct stasis_message_type *history_msg_type(void);
}

CallHistory::CallHistory()
{
	if (!this->sqlite.open(OPERATOR_CallHistoryDb, true, false, true, false)) {
		return;
	}

	if (!this->checkDatabase()) {
		this->sqlite.close();
		return;
	}
}

CallHistory::~CallHistory()
{
	this->sqlite.close();
}

bool CallHistory::createDatabase()
{
	ast_debug(3, "Create CDR database...\n");

	if (!this->sqlite.exec(SQL_CDR)) {
		ast_log(AST_LOG_ERROR, "Failed to create table cdr.\n");
		return false;
	}

	if (!this->sqlite.exec("CREATE INDEX FromNumIdx ON cdr (FromNum)")) {
		ast_log(AST_LOG_ERROR, "Failed to create index FromNumIdx.\n");
		return false;
	}

	if (!this->sqlite.exec("CREATE INDEX ToNumIdx ON cdr (FromNum)")) {
		ast_log(AST_LOG_ERROR, "Failed to create index ToNumIdx.\n");
		return false;
	}

	if (!this->sqlite.exec("CREATE INDEX CallInitializedIdx ON cdr (CallInitialized)")) {
		ast_log(AST_LOG_ERROR, "Failed to create index CallInitializedIdx.\n");
		return false;
	}

	ast_debug(3, "Create CDR database... done.\n");
	return true;
}

bool CallHistory::checkDatabase()
{
	ast_debug(3, "Check CDR database...\n");

	if (this->sqlite.hasTable("cdr")) {
		ast_debug(3, "Check CDR database... cdr table exists.\n");
	}
	else {
		this->createDatabase();
		return true;
	}

	/* upgrades */
	/* add a new field to the database */
	ast_debug(3, "Check field FromExternal...\n");
	if (this->sqlite.hasTableColumn("cdr", "FromExternal")) {
		ast_debug(3, "Check field FromExternal... done.\n");
	}
	else {
		ast_debug(3, "Check field FromExternal... missing.\n");
		/* upgrade */
		ast_debug(3, "Check field FromExternal... upgrade database...\n");
		bool success = true;
		const char *oldFields = "Guid,FromNum,FromName,FromChannel,FromType,ToNum,ToChannel,ToType,Status,External,AccountCode,CallInitialized,CallDuration,AnsweredDuration,UniqueId";
		success = success && this->sqlite.exec("BEGIN TRANSACTION;");
		success = success && this->sqlite.exec("CREATE TEMPORARY TABLE cdr_backup(%s);", oldFields);
		success = success && this->sqlite.exec("INSERT INTO cdr_backup (%s) SELECT %s FROM cdr;", oldFields, oldFields);
		success = success && this->sqlite.exec("DROP TABLE cdr;");
		success = success && createDatabase();
		success = success && this->sqlite.exec("INSERT INTO cdr (%s) SELECT %s FROM cdr_backup;", oldFields, oldFields);
		success = success && this->sqlite.exec("DROP TABLE cdr_backup");
		this->sqlite.exec(success ? "COMMIT;" : "ROLLBACK;");
		this->sqlite.exec("VACUUM");
		ast_debug(3, "Check field FromExternal... upgrade database... done.\n");
		return success;
	}

	ast_debug(3, "Check CDR database... done.\n");
	return true;
}

void CallHistory::updateCallHistoryStatus(KTS_CallHistoryStatus &status, const std::string value)
{
	if ("Busy" == value) {
		status = KTS_CallHistoryStatusBusy;
	}
	else if ("ReceiveFax" == value || "SendFax" == value) {
		/* ignored */
	}
	else if ("Forbidden" == value) {
		status = KTS_CallHistoryStatusForbidden;
	}
	else if ("Voicemail" == value) {
		status = KTS_CallHistoryStatusVoicemail;
	}
	else if ("Followme" == value) {
		status = KTS_CallHistoryStatusFollowme;
	}
	else if ("Failed" == value) {
		status = KTS_CallHistoryStatusFailed;
	}
	else if ("Blacklisted" == value) {
		status = KTS_CallHistoryStatusBlacklisted;
	}
	else {
		ast_log(AST_LOG_WARNING, "Unknown KTS_CallHistoryStatus \"%s\"\n", value.c_str());
	}
}

std::string CallHistory::readCdrVariable(struct ast_cdr *p_cdr, const std::string &name, const std::string defaultValue /* = "" */)
{
	char tmpBuf[1024];
	char *value = NULL;

	ast_cdr_format_var(p_cdr, name.c_str(), &value, tmpBuf, sizeof (tmpBuf), 0);
	if (value && strlen(value)) {
		return std::string(value);
	}
	return defaultValue;
}

void CallHistory::updateExtensionTypes(const std::string fromNum, const std::string toNum, KTS_ExtensionType &fromType, KTS_ExtensionType &toType)
{
	/* update fromType and toType */
	if (fromType == KTS_ExtensionTypeUnknown) {
		fromType = activecalls_getExtensionTypeConf(fromNum);
	}

	if (toType == KTS_ExtensionTypeUnknown) {
		toType = activecalls_getExtensionTypeConf(toNum);
	}
}

std::string CallHistory::getCodecName(const std::string &codecName)
{
	struct ast_codec *codec;
	std::string codecName2(codecName);

	if ((*codecName.begin() == '(') & (*codecName.rbegin() == ')')) {
		codecName2 = codecName.substr(1, codecName.size() - 2);
	}

	if (codecName2.empty() || codecName2 == "nothing" || codecName2 == "none") {
		return "";
	}

	for (int i = 1; (codec = ast_codec_get_by_id(i)); i++) {
		if (codecName2 == codec->name) {
			std::string ret = codec->description;
			ao2_ref(codec, -1);
			return ret;
		}

		ao2_ref(codec, -1);
	}

	return codecName;
}

void CallHistory::amiSendMessage(const std::string &fromNum, const std::string &toNum)
{
	struct ast_json *msg = ast_json_object_create();
	ast_json_object_set(msg, "fromNum", ast_json_string_create(fromNum.c_str()));
	ast_json_object_set(msg, "toNum", ast_json_string_create(toNum.c_str()));

	ami_send_message(history_msg_type(), msg);
	ast_json_unref(msg);
}

/**
 * Function called from Asterisk's cdr module after a call has ended. However,
 * there might be a minor delay, it's not called straight away.
 * Started in a separate thread, so timing is not critical.
 */
int CallHistory::cdrLog(struct ast_cdr * p_cdr)
{
	std::string tmpStr;

	/* FROM */
	std::string fromNum = p_cdr->src;
	std::string fromName = p_cdr->clid;
	std::string fromChannel = p_cdr->channel;
	//	KTS_ExtensionType fromType = KTS_ExtensionTypeUnknown;

	/* TO */
	std::string toChannel = p_cdr->dstchannel;
	std::string toNum = p_cdr->dst;
	//	KTS_ExtensionType toType = KTS_ExtensionTypeUnknown;

	/* DURATION */
	int callInitialized = p_cdr->start.tv_sec;
	int callDuration = p_cdr->duration;
	int answeredDuration = p_cdr->billsec;
	if (p_cdr->end.tv_sec > 0) {
		callDuration = p_cdr->end.tv_sec - p_cdr->start.tv_sec;
	}

	/* STATUS */
	KTS_CallHistoryStatus callStatus;
	switch (p_cdr->disposition) {
		case AST_CDR_FAILED:
			callStatus = KTS_CallHistoryStatusFailed;
			break;
		case AST_CDR_BUSY:
			callStatus = KTS_CallHistoryStatusBusy;
			break;
		case AST_CDR_NOANSWER:
			callStatus = KTS_CallHistoryStatusNoAnswer;
			break;
		case AST_CDR_ANSWERED:
			callStatus = KTS_CallHistoryStatusAnswered;
			if (answeredDuration == 0) {
				answeredDuration = 1; /* billsec is 0 for calls <1s. */
			}
			break;
		case AST_CDR_NULL:
			/* nobreak */
		default:
			callStatus = KTS_CallHistoryStatusUnknown;
	}

	/* CDR fields override everything */
	KTS_ExtensionType fromType = KTS_ExtensionTypeUnknown;
	KTS_ExtensionType toType = KTS_ExtensionTypeUnknown;
	std::string fromExternal, toExternal;

	fromNum = readCdrVariable(p_cdr, "ktsFromInit", fromNum);
	toNum = readCdrVariable(p_cdr, "ktsToInit", toNum);

	tmpStr = readCdrVariable(p_cdr, "ktsStatusInit");
	if (!tmpStr.empty()) {
		updateCallHistoryStatus(callStatus, tmpStr);
	}

	tmpStr = readCdrVariable(p_cdr, "ktsFromTypeInit");
	if (!tmpStr.empty()) {
		fromType = activecalls_getExtensionType(tmpStr);
	}

	tmpStr = readCdrVariable(p_cdr, "ktsToTypeInit");
	if (!tmpStr.empty()) {
		toType = activecalls_getExtensionType(tmpStr);
	}

	fromExternal = readCdrVariable(p_cdr, "ktsFromExternalInit", fromExternal);
	toExternal = readCdrVariable(p_cdr, "ktsToExternalInit", toExternal);

	/* update fromType and toType from kerio database when unknown */
	updateExtensionTypes(fromNum, toNum, fromType, toType);

	/* HERE IS A PLACE FOR SOME HACKS */

	/* skip call parking - weird entries */
	if (fromChannel.c_str() == strstr(fromChannel.c_str(), "Parking/")) {
		ast_debug(3, "Call parking skip (%s).\n", fromChannel.c_str());
		return 0;
	}

	/* call parking */
	if (p_cdr->dst) {
		boost::smatch match;
		std::string parkExten = p_cdr->dst;
		if (boost::regex_search(parkExten, match, boost::regex("^SIP_(.+)$"))) {
			if (match.size() == 2) {
				toNum = match[1];
			}
		}
	}

	/* skip duplicit faxes */
	if (0 == strcmp(p_cdr->dst, "faxSend")) {
		return 0;
	}

	/* skip calls with s as the destination - this is here because we log
	 * unanswered calls to cdr - cdr.conf, option unanswered
	 */
	if (toNum == "s") {
		ast_debug(3, "Call with toNum='s' detected (%s, %s, %s).\n", p_cdr->dst, p_cdr->dstchannel, toNum.c_str());
		return 0;
	}
	/* skip unknown calls */
	if (callStatus == KTS_CallHistoryStatusUnknown) {
		ast_debug(3, "Call with unknown status detected (%s, %s, %s).\n", p_cdr->dst, p_cdr->dstchannel, toNum.c_str());
		return 0;
	}
	/* skip calls without from and to - can happen when sending faxes to a non-existent number */
	if (fromNum.empty() && toNum.empty()) {
		ast_debug(3, "Call with empty from and to detected.\n");
		return 0;
	}

	/* update durations */
	if (KTS_CallHistoryStatusNoAnswer == callStatus) {
		answeredDuration = 0;
	}
	if (KTS_CallHistoryStatusForbidden == callStatus) {
		answeredDuration = 0;
	}
	if (KTS_CallHistoryStatusBlacklisted == callStatus) {
		answeredDuration = 0;
	}

	std::string fromInterface = readCdrVariable(p_cdr, "ktsFromInterface");
	std::string toInterface = readCdrVariable(p_cdr, "ktsToInterface");
	std::string fromPeerIp = readCdrVariable(p_cdr, "ktsFromPeerIp");
	std::string toPeerIp = readCdrVariable(p_cdr, "ktsToPeerIp");
	std::string fromRecvIp = readCdrVariable(p_cdr, "ktsFromRecvIp");
	std::string toRecvIp = readCdrVariable(p_cdr, "ktsToRecvIp");
	std::string fromUserAgent = readCdrVariable(p_cdr, "ktsFromUserAgent");
	std::string toUserAgent = readCdrVariable(p_cdr, "ktsToUserAgent");
	std::string fromCodec = getCodecName(readCdrVariable(p_cdr, "ktsFromCodec"));
	std::string toCodec = getCodecName(readCdrVariable(p_cdr, "ktsToCodec"));
	std::string fromQos = readCdrVariable(p_cdr, "ktsFromQos");
	std::string toQos = readCdrVariable(p_cdr, "ktsToQos");
	std::string details = readCdrVariable(p_cdr, "ktsDetails");
	std::string ktsMyPhoneTo = readCdrVariable(p_cdr, "ktsMyPhoneTo");


	std::string fromCodecVideo = getCodecName(readCdrVariable(p_cdr, "ktsFromCodecVideo"));
	std::string toCodecVideo = getCodecName(readCdrVariable(p_cdr, "ktsToCodecVideo"));
	if (!fromCodecVideo.empty()) {
		fromCodec += ", " + fromCodecVideo;
	}
	if (!toCodecVideo.empty()) {
		toCodec += ", " + toCodecVideo;
	}

	ast_debug(3, "callStatus:  %d\n", (int) callStatus);
	ast_debug(3, "ktsMyPhoneTo:         %s\n", ktsMyPhoneTo.c_str());
	ast_debug(3, "src:         %s\n", p_cdr->src);
	ast_debug(3, "dst:         %s\n", p_cdr->dst);
	ast_debug(3, "channel:     %s\n", p_cdr->channel);
	ast_debug(3, "dstchannel:  %s\n", p_cdr->dstchannel);
	ast_debug(3, "lastapp:     %s\n", p_cdr->lastapp);
	ast_debug(3, "disposition: %ld\n", p_cdr->disposition);
	ast_debug(3, "start:       %d\n", (int)p_cdr->start.tv_sec);
	ast_debug(3, "answer:      %d\n", (int)p_cdr->answer.tv_sec);
	ast_debug(3, "end:         %d\n", (int)p_cdr->end.tv_sec);
	ast_debug(3, "billsec:     %d\n", (int)p_cdr->billsec);

	/* fromName contains: "Admin" <20>, strip it to Admin */
	size_t last = fromName.find_last_of('<');
	if (last != std::string::npos) {
		fromName.erase(last); /* remove <...> */
		boost::algorithm::trim(fromName); /* remove spaces */

		/* remove first '"' */
		if (!fromName.empty() && fromName[0] == '"') {
			fromName.erase(0, 1);
		}

		/* remove last '"' */
		if (!fromName.empty() && fromName[fromName.size() - 1] == '"') {
			fromName.erase(fromName.size() - 1);
		}
	}

	/* webrtc ip addresses */
	static const boost::regex webrtcUserAgent("^" OPERATOR_WebrtcUserAgent "@([0-9\\.]+) ");
	boost::smatch match;
	if (boost::regex_search(fromUserAgent, match, webrtcUserAgent)) {
		fromRecvIp = match[1];
		fromPeerIp = match[1];
	}
	if (boost::regex_search(toUserAgent, match, webrtcUserAgent)) {
		toPeerIp = match[1];
		toPeerIp = match[1];
	}

	/* save call to the database */
	if (!this->sqlite.isOpen()) {
		return 0; /* failure */
	}

	ast_debug(3, "Insert call %s (%s->%s) into CDR.\n", p_cdr->uniqueid, fromChannel.c_str(), toChannel.c_str());
	this->sqlite.exec(
			"INSERT INTO cdr ("
			"FromNum,"
			"FromName,"
			"FromChannel,"
			"FromType,"
			"ToNum,"
			"ToChannel,"
			"ToType,"
			"Status,"
			"External,"
			"AccountCode,"
			"CallInitialized,"
			"CallDuration,"
			"AnsweredDuration,"
			"UniqueId,"
			"FromInterface,"
			"ToInterface,"
			"FromExternal,"
			"ToExternal,"
			"FromPeerIp,"
			"ToPeerIp,"
			"FromRecvIp,"
			"ToRecvIp,"
			"FromUserAgent,"
			"ToUserAgent,"
			"FromCodec,"
			"ToCodec,"
			"FromQos,"
			"ToQos,"
			"Details"
			") VALUES ("
			"'%q', '%q', '%q', %d, "
			"'%q', '%q', %d, %d, "
			"'%q', '%q', %d, %d, %d, "
			"'%q', '%q', '%q', '%q', '%q', '%q', '%q', '%q', '%q', '%q', '%q', '%q', '%q', '%q', '%q', '%q'"
			");",
			fromNum.c_str(),
			fromName.c_str(),
			fromChannel.c_str(),
			(int) fromType,
			toNum.c_str(),
			toChannel.c_str(),
			(int) toType,
			(int) callStatus,
			"", //TODO
			p_cdr->accountcode,
			callInitialized,
			callDuration,
			answeredDuration,
			p_cdr->uniqueid,
			fromInterface.c_str(),
			toInterface.c_str(),
			fromExternal.c_str(),
			toExternal.c_str(),
			fromPeerIp.c_str(),
			toPeerIp.c_str(),
			fromRecvIp.c_str(),
			toRecvIp.c_str(),
			fromUserAgent.c_str(),
			toUserAgent.c_str(),
			fromCodec.c_str(),
			toCodec.c_str(),
			fromQos.c_str(),
			toQos.c_str(),
			details.c_str()
			);

	this->amiSendMessage(fromNum, toNum);
	return 0; /* success */
}

/**
 * Update cdr argument.
 *
 * Updates paramName on every run and paramName+"Init" only once
 */
void CallHistory::asteriskUpdateParam(struct ast_channel *chan, const std::string paramName, const std::string paramValue, bool rewriteInit)
{
	std::string paramNameInit = paramName + "Init";
	const char *channelName = ast_channel_name(chan);
	std::string value;

	/* CHANNEL */
	value = ActiveCalls::getChannelVariable(paramNameInit, chan);
	if ((value.empty() && !paramValue.empty()) || rewriteInit) {
		/* paramNameInit doesn't exist & paramValue is not empty */
		/* or rewriteInit is true */

		ast_debug(3, "Chan update Init %s=%s\n", paramNameInit.c_str(), paramValue.c_str());
		ActiveCalls::setChannelVariable(chan, paramNameInit, paramValue);
	}
	ast_debug(3, "Chan update Init %s=%s\n", paramNameInit.c_str(), paramValue.c_str());
	ActiveCalls::setChannelVariable(chan, paramName, paramValue);


	/* CDR */
	value = ActiveCalls::getCdrVariable(paramNameInit, channelName);
	if ((value.empty() && !paramValue.empty()) || rewriteInit) {
		/* paramNameInit doesn't exist & paramValue is not empty */
		/* or rewriteInit is true */

		ast_debug(3, "CDR  update Init %s=%s\n", paramNameInit.c_str(), paramValue.c_str());
		ast_cdr_setvar(channelName, paramNameInit.c_str(), paramValue.c_str());
	}

	ast_debug(3, "CDR  update      %s=%s\n", paramName.c_str(), paramValue.c_str());
	ast_cdr_setvar(channelName, paramName.c_str(), paramValue.c_str());
}

/**
 *  Update CDR channel variables acording to arguments arg
 */
void CallHistory::asteriskUpdateParams(struct ast_channel *chan, const std::string argStr)
{
	std::vector<std::string> argVect;
	boost::split(argVect, argStr, boost::is_any_of(","));

	MapString celStrings;

	BOOST_FOREACH(std::string argument, argVect)
	{
		boost::smatch match;
		if (boost::regex_search(argument, match, boost::regex("([^=]+)=([^|,]*)"))) {
			if (match.size() == 3) {
				asteriskUpdateParam(chan, match[1].str(), match[2].str(), false);
				celStrings[match[1].str()] = match[2].str();
			}
		}
	}

	/* CEL */
	struct ast_json *celJson = ast_json_pack("{s: s, s: o}",
			"event", "kerio",
			"extra", createJson(&celStrings));
	char *celStr = ast_json_dump_string(celJson);
	ast_debug(3, "Sending custom CEL event: %s\n", celStr);
	ast_cel_publish_event(chan, AST_CEL_USER_DEFINED, celJson);
	ast_json_unref(celJson);
	ast_json_free(celStr);
}

/**
 * Get count of calls with prefixes and a timeframe
 *
 * @param scope [int] - 0:all, 1:only, 2:except
 * @param prefixes [std::vector<std::string>] - vector of prefixes
 * @param timeframe [int] - timeframe
 * @return [int] - count of calls
 */
int CallHistory::getCallsCountPrefix(int scope, std::vector<std::string> prefixes, int timeframe)
{
	std::string sqlQuery;
	std::string glue;
	std::stringstream ss;

	ss << "SELECT COUNT(Guid) FROM cdr WHERE CallInitialized >= " << timeframe
			<< " AND Status = " << KTS_CallHistoryStatusAnswered
			<< " AND ToType = " << KTS_ExtensionTypeOutboundRoute;
	switch (scope) {
		case 0: /* all */
			break;
		case 1: /* only */
			ss << " AND (";

			BOOST_FOREACH(std::string prefix, prefixes)
			{
				ss << glue;
				ss << "toNum LIKE '" << prefix << "%%'";
				glue = " OR ";
			}
			ss << ")";
			break;
		case 2: /* except */
			ss << " AND (";

			BOOST_FOREACH(std::string prefix, prefixes)
			{
				ss << glue;
				ss << "toNum NOT LIKE '" << prefix << "%%'";
				glue = " AND ";
			}
			ss << ")";
			break;
	}

	sqlQuery = ss.str();

	if (!this->sqlite.isOpen()) {
		return 0; /* failure */
	}

	SqliteResult result;
#pragma GCC diagnostic ignored "-Wformat-security"
	if (!this->sqlite.query(result, sqlQuery.c_str())) {
		ast_log(LOG_ERROR, "Failed to get call count.\n");
		return 0; /* failure */
	}
#pragma GCC diagnostic warning "-Wformat-security"

	int count = 0;
	if (this->sqlite.fetchRow(result)) {
		count = sqlite3_column_int(result.statement, 0);
	}
	else {
		ast_log(LOG_ERROR, "Failed to fetch call count.\n");
	}

	ast_debug(3, "Call count \"%s\" = %d\n", sqlQuery.c_str(), count);
	return count > 0 ? count : 0;
}

/**
 * Get count of calls with prefixes and a timeframe
 *
 * @param scope [int] - 0:all, 1:only, 2:except
 * @param prefixes [std::vector<std::string>] - vector of prefixes
 * @param timeframe [int] - timeframe
 * @return [int] - duration of calls
 */
int CallHistory::getCallsDurationPrefix(int scope, std::vector<std::string> prefixes, int timeframe)
{
	std::string sqlQuery;
	std::string glue;
	std::stringstream ss;

	ss << "SELECT Sum(CallDuration) FROM cdr WHERE CallInitialized >= " << timeframe
			<< " AND Status = " << KTS_CallHistoryStatusAnswered
			<< " AND ToType = " << KTS_ExtensionTypeOutboundRoute;
	switch (scope) {
		case 0: /* all */
			break;
		case 1: /* only */
			ss << " AND (";

			BOOST_FOREACH(std::string prefix, prefixes)
			{
				ss << glue;
				ss << "toNum LIKE '" << prefix << "%%'";
				glue = " OR ";
			}
			ss << ")";
			break;
		case 2: /* except */
			ss << " AND (";

			BOOST_FOREACH(std::string prefix, prefixes)
			{
				ss << glue;
				ss << "toNum NOT LIKE '" << prefix << "%%'";
				glue = " AND ";
			}
			ss << ")";
			break;
	}

	sqlQuery = ss.str();

	if (!this->sqlite.isOpen()) {
		return 0; /* failure */
	}

	SqliteResult result;
#pragma GCC diagnostic ignored "-Wformat-security"
	if (!this->sqlite.query(result, sqlQuery.c_str())) {
		ast_log(LOG_ERROR, "Failed to get call duration.\n");
		return 0; /* failure */
	}
#pragma GCC diagnostic warning "-Wformat-security"

	int duration = 0;
	if (this->sqlite.fetchRow(result)) {
		duration = sqlite3_column_int(result.statement, 0);
	}
	else {
		ast_log(LOG_ERROR, "Failed to fetch call duration.\n");
	}

	ast_debug(3, "Call duration \"%s\" = %d\n", sqlQuery.c_str(), duration);
	return duration > 0 ? duration : 0;
}
