#include <boostWrapper.h>
#include <boost/regex.hpp>

#include "Common.h"

#ifndef AST_MODULE
#define AST_MODULE "Kerio"
#endif

/* same as mkstemp, but return a FILE * */
FILE *vm_mkftemp(char *tmpl)
{
	FILE *p = NULL;
	int pfd = mkstemp(tmpl);
	if (pfd > -1) {
		p = fdopen(pfd, "w+");
		if (!p) {
			close(pfd);
			pfd = -1;
		}
	}
	return p;
}

std::string getExtFromChannel(const std::string &channel)
{
	boost::smatch match;

	/* SIP, example: SIP/1234-081d9940 */
	if (boost::regex_search(channel, match, boost::regex("SIP/([0-9]+)"))) {
		if (match.size() == 2) {
			return match[1].str();
		}
	}

	return "";
}

bool shellExec(const std::string cmd, std::string &output)
{
	FILE *fp;
	char tmpBuff[BUFSIZ];

	ast_debug(2, "Executing: %s...\n", cmd.c_str());
	fp = popen(cmd.c_str(), "r");
	if (!fp) {
		ast_log(AST_LOG_WARNING, "shellExec: Failed to execute command '%s'\n", cmd.c_str());
		return false; /* error */
	}

	while (fgets(tmpBuff, sizeof (tmpBuff), fp)) {
		output += tmpBuff;
	}
	pclose(fp);

	ast_debug(3, "Executing: %s... Done[%s]\n", cmd.c_str(), output.c_str());
	return true; /* OK */
}

bool fileChanged(const char *filename, int &lastTimestamp)
{
	/* check timestamp */
	struct stat localStat;
	if (0 != stat(filename, &localStat)) {
		ast_log(AST_LOG_ERROR, "File %s doesn't exist\n", filename);
		return false;
	}
	if (lastTimestamp == localStat.st_mtime) {
		return false; /* no change */
	}
	lastTimestamp = localStat.st_mtime;
	return true;
}

bool isNumberValid(const std::string number)
{
	boost::smatch match;
	/* Invalid numbers: "", "h", "s", "t" */
	if (boost::regex_search(number, match, boost::regex("^$|^[^0-9\\+\\#\\*]$"))) {
		return false;
	}
	return true;
}

bool parseJson(const char *jsonString, MapString *outStr, MapInt *outInt, MapBool *outBool)
{
	struct ast_json_error error;

	struct ast_json *json = ast_json_load_string(jsonString, &error);
	if (!json) {
		ast_log(AST_LOG_ERROR, "Json error %d:%d:%d: %s\n", error.line, error.column, error.position, error.text);
		return false;
	}

	for (struct ast_json_iter *iter = ast_json_object_iter(json); iter; iter = ast_json_object_iter_next(json, iter)) {
		std::string name = ast_json_object_iter_key(iter);
		ast_json *value = ast_json_object_iter_value(iter);

		switch (ast_json_typeof(value)) {
			case AST_JSON_STRING:
				if (outStr) {
					(*outStr)[name] = ast_json_string_get(value);
				}
				break;
			case AST_JSON_INTEGER:
				if (outInt) {
					(*outInt)[name] = ast_json_integer_get(value);
				}
				break;
			case AST_JSON_TRUE:
				if (outBool) {
					(*outBool)[name] = true;
				}
				break;
			case AST_JSON_FALSE:
				if (outBool) {
					(*outBool)[name] = false;
				}
				break;
			default:
				/* ignore non-trivial types */
				break;
		}
	}
	ast_json_unref(json);
	return true;
}

struct ast_json *createJson(const MapString *inStr, const MapInt *inInt, const MapBool *inBool)
{
	struct ast_json *json = ast_json_object_create();

	if (inStr) {

		BOOST_FOREACH(const MapString::value_type &item, *inStr)
		{
			ast_json_object_set(json, item.first.c_str(), ast_json_string_create(item.second.c_str()));
		}
	}

	if (inInt) {

		BOOST_FOREACH(const MapInt::value_type &item, *inInt)
		{
			ast_json_object_set(json, item.first.c_str(), ast_json_integer_create(item.second));
		}
	}

	if (inBool) {

		BOOST_FOREACH(const MapBool::value_type item, *inBool)
		{
			ast_json_object_set(json, item.first.c_str(), ast_json_boolean(item.second));
		}
	}

	return json; /* do not forget to ast_json_unref */
}

std::string createJsonString(const MapString *inStr, const MapInt *inInt, const MapBool *inBool)
{
	struct ast_json *json = createJson(inStr, inInt, inBool);
	char *str = ast_json_dump_string_format(json, AST_JSON_PRETTY);
	std::string ret = str;
	ast_json_free(str);
	ast_json_unref(json);
	return ret;
}

struct timeval ast_tvmin(struct timeval a, struct timeval b)
{
	if (ast_tvzero(a)) {
		return b;
	}
	if (ast_tvzero(b)) {

		return a;
	}
	return (ast_tvcmp(a, b) < 0) ? a : b;
}

#define TRY_LOCK_REPEAT 10
#define TRY_LOCK_SLEEP_US 10

bool AsteriskChannelLock::lock()
{
	if (this->locked) {
		return true;
	}

	for (int i = 1; i <= TRY_LOCK_REPEAT; ++i) {
		ast_debug(5, "Lock channel '%s'... attempt %d...\n", this->name.c_str(), i);
		if (0 == ast_channel_trylock(this->c)) {
			ast_debug(5, "Lock channel '%s'... attempt %d... locked.\n", this->name.c_str(), i);
			this->locked = true;
			return true;
		}
		sched_yield();
		usleep(TRY_LOCK_SLEEP_US);
	}

	ast_debug(5, "Lock channel '%s'... failed.\n", this->name.c_str());
	return false;
}

void AsteriskChannelLock::unlock()
{
	if (this->locked) {
		ast_debug(5, "Unlock channel '%s'.\n", this->name.c_str());
		ast_channel_unlock(this->c);
		this->locked = false;
	}
}

AsteriskChannelLock::~AsteriskChannelLock()
{
	this->unlock();
}

AsteriskChannelLock::AsteriskChannelLock(struct ast_channel *c) : c(c), locked(false)
{
	this->name = ast_channel_name(c);
}
