#ifndef INTERFACE_H
#define	INTERFACE_H

#include <vector>
#include <boostWrapper.h>
#include <boost/foreach.hpp>

#ifndef AST_MODULE
#define AST_MODULE "kerio"
#endif

extern "C" {
#include "module.h"
	extern ast_rwlock_t kerio_shutdown_lock; // Bug 87674 - Boost crashes asterisk in long plan
}

class Mutex {
public:
	ast_mutex_t m;
	const char *name;

	Mutex(const char *name = "") : name(name)
	{
		ast_mutex_init(&m);
	}
};

class MutexLock {
	Mutex &mutex;
	const char *file;
	const int line;
public:

	MutexLock(Mutex &mutex, const char *file = "", const int line = 0) : mutex(mutex), file(file), line(line)
	{
		ast_debug(10, "Lock %s - %s:%d...\n", mutex.name, file, line);

		ast_mutex_lock(&mutex.m);

		ast_debug(10, "Lock %s - %s:%d... done.\n", mutex.name, file, line);
	}

	~MutexLock()
	{
		ast_debug(10, "Unlocked %s - %s:%d.\n", mutex.name, file, line);
		ast_mutex_unlock(&mutex.m);
	}
};

/**
 * These locks protect each module's configuration (the global object, e.g.
 * 'CallConstraints') so that the module doesn't try to access it from another
 * thread while it's being (un)loaded. Asterisk has no protection against such
 * cases.
 *
 * Warning:
 * LockUnique - nonrecursive! LockUnique can't be locked again when the same Lock is already locked.
 * LockShared - can be locked multiple times
 */
class Lock {
public:
	ast_rwlock_t l;

	Lock()
	{
		ast_rwlock_init(&l);
	}

	~Lock()
	{
		// XXX Bug 81300 - We cannot destroy the lock as other threads are running concurrently
		//                 with the global destructors and might try to use it.
		//
		// WARN: If the destructor is called from exit(), the resources will be freed by the kernel
		// upon process termination, no problem. If it's called from a dlclose() after
		// a `module unload', this presents a *MEMORY LEAK*. However, we don't unload kerio
		// modules anywhere in Operator, so this shouldn't matter.
		//
		//    ast_rwlock_destroy(&l);
	}
};

class LockUnique {
	ast_rwlock_t *l;
public:
	bool ok;

	LockUnique(ast_rwlock_t *l) : l(l), ok(false)
	{
		// The only case when this fails is when asterisk is in the progress of shutting down. See Bug 87674 for details.
		this->ok = (ast_rwlock_tryrdlock(&kerio_shutdown_lock) == 0);
		if (!this->ok) {
			return;
		}
		ast_rwlock_wrlock(l);
	}

	~LockUnique()
	{
		if (this->ok) {
			ast_rwlock_unlock(l);
			ast_rwlock_unlock(&kerio_shutdown_lock);
		}
	}
};

class LockShared {
	ast_rwlock_t *l;
public:

	LockShared(ast_rwlock_t *l) : l(l)
	{
		ast_rwlock_rdlock(&kerio_shutdown_lock);
		ast_rwlock_rdlock(l);
	}

	~LockShared()
	{
		ast_rwlock_unlock(l);
		ast_rwlock_unlock(&kerio_shutdown_lock);
	}
};
#endif	/* INTERFACE_H */

