#include "Routing.h"

#define CONFIG_FILE "/etc/asterisk/kerio_routing.conf"

Routing::Routing()
{
	/* load data */
	try {
		boost::property_tree::ptree pt;
		boost::property_tree::json_parser::read_json("/etc/asterisk/kerio_routing.conf", pt);

		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_routing.incoming"))
		{
			RoutingTableItem rti(val.second);
			routingTableIncoming[rti.iface].push_back(rti);
		}

		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_routing.outgoing"))
		{
			RoutingTableItem rti(val.second);
			if (rti.external.empty()) {
				rti.external = "anonymous";
			}
			routingTableOutgoing[rti.id].push_back(rti);
		}
	}
	catch (std::exception & e) {
		ast_log(AST_LOG_ERROR, "Configuration error %s\n", e.what());
	}

	ast_verb(3, "Loaded %zu incoming interfaces and %zu outgoing routes.\n", routingTableIncoming.size(), routingTableOutgoing.size());
}

Routing::~Routing() { }

void Routing::print(const int fd)
{
	ast_cli(fd, "Routing - incoming\n");
	ast_cli(fd, "------------------\n");

	BOOST_FOREACH(RoutingTableIncoming::value_type route, routingTableIncoming)
	{

		BOOST_FOREACH(RoutingTableItem item, route.second)
		{
			ast_cli(fd, "%s: %s => %s\n", item.iface.c_str(), item.external.c_str(), item.internal.c_str());
		}
	}

	ast_cli(fd, "\n\n");
	ast_cli(fd, "Routing - outgoing\n");
	ast_cli(fd, "------------------\n");

	BOOST_FOREACH(RoutingTableOutgoing::value_type route, routingTableOutgoing)
	{

		BOOST_FOREACH(RoutingTableItem item, route.second)
		{
			ast_cli(fd, "%5d(%10s): %5s => %16s (%s)\n", item.id, item.iface.c_str(), item.internal.c_str(), item.external.c_str(), item.hideCallerId ? "hidden" : "visible");
		}
	}
}

std::string Routing::find_incoming(const char *interface, const char *number)
{
	if (routingTableIncoming.find(interface) == routingTableIncoming.end()) {
		return ""; /* interface not found */
	}

	BOOST_FOREACH(RoutingTableItem item, routingTableIncoming[interface])
	{
		if (item.external == number) {
			return item.internal;
		}
	}
	return ""; /* number not found */
}

std::string Routing::find_incoming_rev(const char *interface, const char *number)
{
	if (routingTableIncoming.find(interface) == routingTableIncoming.end()) {
		return ""; /* interface not found */
	}

	BOOST_FOREACH(RoutingTableItem item, routingTableIncoming[interface])
	{
		if (item.internal == number) {
			return item.external;
		}
	}
	return ""; /* number not found */
}

std::string Routing::find_outgoing(const char *routeId, const char *number)
{
	int routeIdInt = atoi(routeId);

	if (routingTableOutgoing.find(routeIdInt) == routingTableOutgoing.end()) {
		return ""; /* routeId not found */
	}

	BOOST_FOREACH(RoutingTableItem item, routingTableOutgoing[routeIdInt])
	{
		if (item.internal == number) {
			return item.external;
		}
	}
	return ""; /* number not found */
}

bool Routing::is_hidden(const char *routeId, const char *number)
{
	int routeIdInt = atoi(routeId);

	if (routingTableOutgoing.find(routeIdInt) == routingTableOutgoing.end()) {
		return false; /* routeId not found */
	}

	BOOST_FOREACH(RoutingTableItem item, routingTableOutgoing[routeIdInt])
	{
		if (item.internal == number) {
			return item.hideCallerId;
		}
	}

	return false; /* number not found */
}