#include "Interface.h"
#include "TimeRanges.h"
#include "TimeRangeInterface.h"

static TimeRanges *timeRanges;
static Lock loadLock;

int timerange_load()
{
	LockUnique l(&loadLock.l); if (!l.ok) return 0;

	timeRanges = new TimeRanges();
	return 0;
}

void timerange_unload()
{
	LockUnique l(&loadLock.l); if (!l.ok) return;

	if (timeRanges) {
		delete timeRanges;
		timeRanges = NULL;
	}
}

void timerange_print(const int fd)
{
	LockShared l(&loadLock.l);
	if (!timeRanges) {
		return;
	}

	ast_cli(fd, "------------------------------------\n");

	BOOST_FOREACH(TimeRangeEntry te, timeRanges->timeRangeEntryList)
	{
		time_t timeStamp;
		tm *time;

		ast_cli(fd, "groupName:        \"%s\"\n", te.groupName.c_str());
		ast_cli(fd, "entryEnabled:     %d\n", te.entryEnabled);
		ast_cli(fd, "entryDescription: \"%s\"\n", te.entryDescription.c_str());

		switch (te.entryType) {
			case 0: /* daily */
				ast_cli(fd, "entryType:        %d (daily)\n", te.entryType);

				ast_cli(fd, "entryDays:        \"%s\"\n", te.entryDays);
				ast_cli(fd, "entryStart:       %10d (%02d:%02d)\n", te.entryStart, te.entryStart / 3600, (te.entryStart % 3600) / 60);
				ast_cli(fd, "entryEnd:         %10d (%02d:%02d)\n", te.entryEnd, te.entryEnd / 3600, (te.entryEnd % 3600) / 60);

				break;
			case 1: /* weekly */
				ast_cli(fd, "entryType:        %d (weekly)\n", te.entryType);

				timeStamp = te.entryStart;
				time = gmtime(&timeStamp);
				ast_cli(fd, "entryStart:       %10d (day %d %02d:%02d:%02d)\n", te.entryStart,
						time->tm_mday - 1, time->tm_hour, time->tm_min, time->tm_sec);

				timeStamp = te.entryEnd;
				time = gmtime(&timeStamp);
				ast_cli(fd, "entryEnd:         %10d (day %d %02d:%02d:%02d)\n", te.entryEnd,
						time->tm_mday - 1, time->tm_hour, time->tm_min, time->tm_sec);
				break;
			case 2: /* absolute */
				ast_cli(fd, "entryType:        %d (absolute)\n", te.entryType);

				timeStamp = te.entryStart;
				time = gmtime(&timeStamp);
				ast_cli(fd, "entryStart:       %10d (%02d-%02d-%d %02d:%02d:%02d)\n", te.entryStart,
						time->tm_mday, time->tm_mon + 1, time->tm_year + 1900, time->tm_hour, time->tm_min, time->tm_sec);

				timeStamp = te.entryEnd;
				time = gmtime(&timeStamp);
				ast_cli(fd, "entryEnd:         %10d (%02d-%02d-%d %02d:%02d:%02d)\n", te.entryEnd,
						time->tm_mday, time->tm_mon + 1, time->tm_year + 1900, time->tm_hour, time->tm_min, time->tm_sec);

				break;
		}
	}
	ast_cli(fd, "------------------------------------\n");
	ast_cli(fd, "Item count: %zu\n", timeRanges->timeRangeEntryList.size());
}

int timerange_test(const char *timeRange)
{
	LockShared l(&loadLock.l);
	if (!timeRanges) {
		return -1;
	}

	return timeRanges->testCurrentTime(timeRange);
}

int timerange_test_timestamp(const char *timeRange, time_t timeStamp)
{
	LockShared l(&loadLock.l);
	if (!timeRanges) {
		return -1;
	}
	return timeRanges->testTime(timeRange, timeStamp);
}

/**
 * convert std::vector to char**
 * !!!FREE MEMORY!!!
 *
 * @return [const char **]
 */
char **timerange_get_names(void)
{
	LockShared l(&loadLock.l);
	if (!timeRanges) {
		return NULL;
	}

	std::vector<std::string> groupNames = timeRanges->getGroupNames();
	int size = groupNames.size();

	char **result = (char**)malloc((size + 1) * sizeof (char *));

	for (int counter = 0; counter < size; counter++) {
		if (std::string::npos != groupNames[counter].find(' ')) {
			groupNames[counter] = "\"" + groupNames[counter] + "\"";
		}
		result[counter] = strdup(groupNames[counter].c_str());
	}
	result[size] = NULL;

	return result;
}
