#include "TimeRanges.h"

TimeRanges::TimeRanges()
{
	try {
		boost::property_tree::ptree pt;
		boost::property_tree::json_parser::read_json("/etc/asterisk/kerio_timeRanges.conf", pt);

		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_timeRanges.timeRanges"))
		{
			TimeRangeEntry te(val.second);
			timeRangeEntryList.push_back(te);
		}
	}
	catch (std::exception & e) {
		ast_log(AST_LOG_ERROR, "Configuration error %s\n", e.what());
	}
}

TimeRanges::~TimeRanges() { }

/**
 * Get group names
 *
 * @return [std::set<std::string>] - time range group names
 */
std::vector<std::string> TimeRanges::getGroupNames()
{
	std::vector<std::string> result;
	std::set<std::string> tmpSet;

	/* save group names into a vector */
	BOOST_FOREACH(TimeRangeEntry te, timeRangeEntryList)
	{
		tmpSet.insert(te.groupName);
	}

	std::copy(tmpSet.begin(), tmpSet.end(), std::back_inserter(result));
	return result;
}

/**
 * Test time
 *
 * @param timeRangeGroupName [std::string] - time range group name
 * @param timeStamp [time_t] - time to check
 * @return [int] - -1 on error, 0 on miss, 1 on hit
 */
int TimeRanges::testTime(const std::string timeRangeGroupName, time_t timeStamp)
{
	int groupNameFound = 0;
	struct tm *currentTime;
	currentTime = gmtime(&timeStamp);
	currentTime->tm_sec = 0;
	currentTime->tm_min = 0;
	currentTime->tm_hour = 0;

	BOOST_FOREACH(TimeRangeEntry te, timeRangeEntryList)
	{
		if (timeRangeGroupName == te.groupName) {
			groupNameFound = 1;
			if (te.entryEnabled) {
				int skipTimeCheck = 0;
				int timeRangeStart = 0;
				switch (te.entryType) {
					case 0: // daily
						// check day of week
						if (0 == currentTime->tm_wday) {
							if (!strchr(te.entryDays, '7')) {
								skipTimeCheck = 1;
							}
						}
						else if (!strchr(te.entryDays, currentTime->tm_wday + '0')) {
							skipTimeCheck = 1;
						}

						timeRangeStart = timegm(currentTime);

						break;
					case 1: // weekly
						currentTime->tm_mday -= (currentTime->tm_wday == 0) ? 6 : (currentTime->tm_wday - 1);
						timeRangeStart = timegm(currentTime);
						break;
					case 2: // absolute
						timeRangeStart = 0;
						break;
					case 3: // nested group
						// not implemented
						timeRangeStart = 0;
						skipTimeCheck = 1;
						break;
					default: return -1; /* Error */
				}

				if (!skipTimeCheck) {
					/*
					std::cout << "entryEnd             " << te.entryEnd << std::endl;
					std::cout << "entryStart           " << te.entryStart << std::endl;
					std::cout << "timeStamp              " << timeStamp << std::endl;
					std::cout << "timeRangeStart     " << timeRangeStart << std::endl;
					std::cout << "timeRangeStart+end " << ((long) (timeRangeStart + te.entryEnd)) << std::endl;
					std::cout << "timeRangeStart+sta " << ((long) (timeRangeStart + te.entryStart)) << std::endl;
					std::cout << (timeStamp < (long) (timeRangeStart + te.entryEnd)) << std::endl;
					std::cout << (timeStamp >= (long) (timeRangeStart + te.entryStart)) << std::endl;
					 */
					// check time in day
					if (te.entryEnd > te.entryStart) {
						if ((timeStamp <= (long) (timeRangeStart + te.entryEnd)) && (timeStamp >= (long) (timeRangeStart + te.entryStart)))
							return 1;
					}
					else {
						if ((timeStamp <= (long) (timeRangeStart + te.entryEnd)) || (timeStamp >= (long) (timeRangeStart + te.entryStart)))
							return 1;
					}
				}
			}
		}
	}

	if (groupNameFound) {
		return 0; /* Miss */
	}
	else {
		return -1; /* Error */
	}
}

/**
 * Test current time
 *
 * @param timeRangeGroupName [std::string]- name of the time range group
 * @return [int] - -1 on error, 0 on miss, 1 on hit
 */
int TimeRanges::testCurrentTime(const std::string timeRangeGroupName)
{
	time_t gmtTime;
	time(&gmtTime); /* current gmt */

	return testTime(timeRangeGroupName, timegm(localtime(&gmtTime))); /* call with localtime */
}

